package org.krasnow.cng.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.krasnow.cng.domain.LinkedStack;
import org.krasnow.cng.domain.SwcTreeNode;
import org.krasnow.cng.utils.SwcDataUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Between any two nodes, takes Z values at those nodes and interpolates the Z values
 * at continuations in between based on XY path length between continuations.
 * 
 */
public class InterpolateZValues {

	public static List interpolateZ(String swcFilename)throws Exception{
		return interpolateZ(new File(swcFilename));
	}

	public static List interpolateZ(File swcFile) throws Exception{
		List trees = ReadSWC.convertSwcToBinaryTreeList(swcFile);
		for (int i = 0; i < trees.size(); i++)
			interpolateZ((SwcTreeNode)trees.get(i));
		return trees;
	}
	
	public static void interpolateZ(SwcTreeNode root){
		SwcTreeNode node;
		LinkedStack stack = new LinkedStack();
		for (int i = 0; i < root.getChildren().size(); i++){
			stack.push(root.getChildren().get(i));
		}
		while (!stack.isEmpty()){
			node = (SwcTreeNode)stack.pop();
			if (node.getChildren().size() > 1){
				//System.out.println(node.getNodeId());
				for (int i = 0; i < node.getChildren().size(); i++){
					//System.out.println(node.getSwcChild(i).getNodeId());
					stack.push(node.getChildren().get(i));
				}
			}
			else{
				// If the first next node is not a termination, interpolate
				if (node.hasChildren()){
					node = interpolateSegment(node);
				}
				// If the next node is not a termination
				if (node.hasChildren()){
					stack.push(node);
				}
			}
		}
		
	}
	
	private static SwcTreeNode interpolateSegment(SwcTreeNode start){
		SwcTreeNode point = start, parent;
		double dist, xyPath = 0;
		List distances = new ArrayList();
		do {
			parent = (SwcTreeNode)point.getParent();
			dist = SwcDataUtils.getXYDistance(parent.getSwcData(), point.getSwcData());
			xyPath += dist;
			distances.add(new Double(dist));
			point = point.getSwcChild(0);
		} while (point.hasChildren() && point.getChildren().size() == 1);
		// Get distance from last point to next bifurcation or termination
		parent = (SwcTreeNode)point.getParent();
		dist = SwcDataUtils.getXYDistance(parent.getSwcData(), point.getSwcData());
		xyPath += dist;
		
		double zStart = start.getSwcParent().getSwcData().getZ();
		double zEnd = point.getSwcData().getZ();
		double totalZDist = zEnd - zStart;

		point = start;
		int i = 0;
		double newZ;
		do {
			dist = ((Double)distances.get(i)).doubleValue();
			newZ = zStart + (dist / xyPath) * totalZDist;
			System.out.println(point.getNodeId()+" " + dist+" " +newZ);
			point.getSwcData().setZ(newZ);
			point = point.getSwcChild(0);
			i++;
			zStart = newZ;
			// Continue while the point is a path point (not a bifurcation or termination) 
		} while (point.hasChildren() && point.getChildren().size() == 1);
		
		parent = (SwcTreeNode)point.getParent();
		dist = SwcDataUtils.getXYDistance(parent.getSwcData(), point.getSwcData());
		newZ = zStart + (dist / xyPath) * totalZDist;
		
		// Return bifurcation node
		return point;
	}
	
	public static void interpolateZAndWrite(File inputFile, File outputFile) throws Exception{
		String inputFilename = inputFile.getName();
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(2);
		if (inputFile.exists()){

			// Read file to preserve header comments
			StringBuffer sb = new StringBuffer();
			BufferedReader reader = new BufferedReader(new FileReader(inputFile));
			String dataString;
			while ( (dataString=reader.readLine()) != null ) { 
				if (dataString.startsWith("#")){
					sb.append(dataString+"\n");
				}
			}
			
			// Interpolate Z values between nodes
			SwcTreeNode root, node;
			List rootList = interpolateZ(inputFile);
			// Start stack to traverse nodes
			LinkedStack stack = new LinkedStack();
			for (int i = 0; i < rootList.size(); i++){
				root = (SwcTreeNode)rootList.get(i);
				stack.push(root);
				
				List nodeList = new ArrayList();
				// Generate list of nodes and sort by id
				while (!stack.isEmpty()){
					node = (SwcTreeNode)stack.pop();
					nodeList.add(node);
	
					for (int j = node.getChildren().size() - 1; j >= 0; j--){
						stack.push(node.getSwcChild(j));
					}
				}
				Collections.sort(nodeList);
				
				
				for (int j = 0; j < nodeList.size(); j++){
					node = (SwcTreeNode)nodeList.get(j);
					sb.append(node.getNodeId()+" "+node.getType() 
							+ " "+node.getX()+" "+node.getY()+" "+nf.format(node.getZ())
							+ " "+node.getRadius()+" "+node.getParentId()+"\n");
				}
	
			}
			DataWriter.writeStringToFile(outputFile, sb.toString());
		}
		else{
			System.out.println("inputFile "+inputFilename+" does not exist");
		}
	}
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			String filename = config.getString("input");
			String outFilename = config.getString("output");
			
			File inputFile = new File(filename), outputFile = new File(outFilename);
			if (inputFile.isDirectory()){
				if (!outputFile.isDirectory()){
					outputFile = new File(inputFile.getAbsolutePath()+"/Z-Interpolated/");
					if (!outputFile.exists()){
						outputFile.mkdir();
					}
				}
				File[] files = inputFile.listFiles();
				for (int i = 0; i < files.length; i++){
					inputFile = files[i];
					if (!inputFile.isDirectory()&&
							(inputFile.getName().endsWith(".swc")||inputFile.getName().endsWith(".SWC"))){
						interpolateZAndWrite(inputFile, new File(outputFile.getAbsolutePath()+"/"+inputFile.getName()));
					}
				}
			}
			else{
				interpolateZAndWrite(inputFile, outputFile);
			}
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption("input")
            .setStringParser(JSAP.STRING_PARSER)
            .setRequired(true)
            .setShortFlag('i') 
            .setLongFlag("input");
        jsap.registerParameter(opt);

        opt = new FlaggedOption("output")
	        .setStringParser(JSAP.STRING_PARSER)
	        .setRequired(true)
	        .setShortFlag('o') 
	        .setLongFlag("output");
	    jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + InterpolateZValues.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

	
}