package org.krasnow.cng.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.DataFormatException;

import org.krasnow.cng.domain.BinaryTreeNode;
import org.krasnow.cng.domain.EuclideanPoint;
import org.krasnow.cng.domain.LinkedStack;
import org.krasnow.cng.domain.NodePair;
import org.krasnow.cng.domain.SwcDataNode;
import org.krasnow.cng.domain.SwcSecondaryData;
import org.krasnow.cng.domain.SwcTreeNode;
import org.krasnow.cng.domain.TreeNode;
import org.krasnow.cng.utils.SwcDataUtils;

/**
 * 
 * @author gillette
 * Converts .swc files info .fasta files and appends them all together into one database.
 * Conversion based on method of division (full cell, arborizations, trees), primary 
 * method of traversal (height, size), and order of traversal (small then large, large
 * then small).  
 * 
 */
public class ReadSWC {
	
	private static boolean debug = false;
	public static double TRAJECTORY_NONE = Math.PI;

	/**
	 * 
	 * @param inputFile
	 * @param zInPathDist
	 * @param scaleZ
	 * @param trajectoryThresholds
	 * @return List of SwcTreeNode trees, each with a separate root (parentId = -1)
	 * @throws Exception
	 */
	public static List<SwcTreeNode> convertSwcToSwcTreeList(
			File inputFile, double scaleZ) throws Exception{

		List<SwcTreeNode> trees = new ArrayList<SwcTreeNode>();
		BufferedReader reader = null;
		
		try{
			reader = new BufferedReader(new FileReader(inputFile));

			String dataString = "";
			Map<Integer,SwcTreeNode> nodeMap = new HashMap<Integer,SwcTreeNode>();
			Map<SwcTreeNode,Integer> addNodeSideMap = new HashMap<SwcTreeNode,Integer>();
			Set<Integer> nodeIds = new TreeSet<Integer>();
			SwcTreeNode node, parent;
			Pattern p = Pattern.compile("[ \\t]*(\\d+)[ \\t]+(\\d+)[ \\t]+(\\-?[\\.\\d]+)[ \\t]+(\\-?[\\.\\d]+)[ \\t]+(\\-?[\\.\\d]+)[ \\t]+([\\.\\d]+)[ \\t]+(\\-?\\d+)[ \\t]*");
			Matcher m;
			int nodeId, parentId, line = 0;
			
			// 1. Assembles entries in swc file, each getting its own SwcTreeNode with any number of children
			while ( (dataString=reader.readLine()) != null ) {
				line++;
		        m = p.matcher(dataString);
		        //System.out.println(dataString);
		        
		        if (m.matches()){
		        	node = new SwcTreeNode();
		        	nodeId = Integer.parseInt(m.group(1));
		        	if (nodeIds.contains(new Integer(nodeId))){
		        		throw new DataFormatException("Node id "+nodeId+" already exists. Second instance found on line "+line+" of "+inputFile.getName());
		        	}
		        	nodeIds.add(nodeId);
		        	node.setNodeId(nodeId);
		        	//type = Integer.parseInt(m.group(2));
		        	//node.setType(type);
		        	// Node types will all be set to 1
		        	node.setType(1);
		        	node.setX(Double.parseDouble(m.group(3)));
		        	node.setY(Double.parseDouble(m.group(4)));
		        	if (scaleZ == 1){
		        		node.setZ(Double.parseDouble(m.group(5)));
		        	}
		        	else{
		        		node.setZ(scaleZ * Double.parseDouble(m.group(5)));
		        	}
		        	node.setRadius(Double.parseDouble(m.group(6)));
		        	parentId = Integer.parseInt(m.group(7));
		        	node.setParentId(parentId);
		        	
		        	// If parentId is -1, this is a root node, so create a new tree rooted at this point
		        	if (parentId == -1){
		        		trees.add(node);
		        	}
		        	// Require that the node referenced by the parent id has been parsed
		        	else if (!nodeIds.contains(new Integer(parentId))){
		        		throw new DataFormatException("Error on line "+line+" of "+inputFile.getName()+". Node "+nodeId+" specifies parent id "+parentId+" which is not in this file or has not yet been defined. A parent must be defined before any of its children.");
		        	}
		
	        		// Get parent from nodeMap
		        	parent = null;
		        	if (nodeMap.containsKey(new Integer(parentId))){
		        		parent = (SwcTreeNode)nodeMap.get(new Integer(parentId));
		        	}
	
	    			addNode(node,parent,nodeMap,addNodeSideMap);
	        	}
		        else if (dataString.length() > 0 && dataString.charAt(0) != '#'){
		        	throw new DataFormatException("Improper SWC format at line "+line+" of "+inputFile.getName());
		        }
		        m = null;
		    }
		} catch (Exception e){
			e.printStackTrace();
		} finally{
			if (reader != null){
				reader.close();
				reader = null;
			}
		}
		
		return trees;
	}

	/**
	 * 
	 * @param inputFile
	 * @param classification
	 * @param minSize
	 * @param calcPathDist
	 * @return
	 * List of SwcTreeNode with child list of BinaryTreeNode. 
	 * Assumes any number of binary trees branching from each root,
	 * regardless of type. Generally there will just be one tree per root 
	 * (a single child of each root), and just one root (one root node in the file).
	 * @throws Exception
	 */
	public static List<SwcTreeNode> convertSwcToBinaryTreeList(
			File inputFile) throws Exception{
		return convertSwcToBinaryTreeList(inputFile, true, 1);
	}
	public static List<SwcTreeNode> convertSwcToBinaryTreeList(File inputFile, boolean zInPathDist) throws Exception{
		return convertSwcToBinaryTreeList(inputFile, zInPathDist, 1, null);
	}
	public static List<SwcTreeNode> convertSwcToBinaryTreeList(File inputFile, double scaleZ) throws Exception{
		return convertSwcToBinaryTreeList(inputFile, true, scaleZ, null);
	}
	
	public static List<SwcTreeNode> convertSwcToBinaryTreeList(
			File inputFile, boolean zInPathDist, double scaleZ) throws Exception{
		return convertSwcToBinaryTreeList(inputFile, zInPathDist, scaleZ, null);
	}
	public static List<SwcTreeNode> convertSwcToBinaryTreeList(
			File inputFile, boolean zInPathDist, double scaleZ, 
			double xyTrajectoryThreshold, double zTrajectoryThreshold) throws Exception{
		SwcDataNode tmp = new SwcDataNode();
		tmp.setX(xyTrajectoryThreshold);
		tmp.setZ(zTrajectoryThreshold);
		return convertSwcToBinaryTreeList(inputFile, zInPathDist, scaleZ, tmp);
		
	}
	public static List<SwcTreeNode> convertSwcToBinaryTreeList(
			File inputFile, boolean zInPathDist, double scaleZ, SwcDataNode trajectoryThresholds) throws Exception{
	    
		// 1. Builds individual trees
		// 2. Removes continuations.
		// 3. Transforms each into binary tree since all tree info is now known (provides stability in transformation).

		List<SwcTreeNode> trees = convertSwcToSwcTreeList(inputFile, scaleZ);

		SwcTreeNode root;
		
		BinaryTreeNode binaryTree;
		List<BinaryTreeNode> binaryTrees;
		// Go through each tree
		for (int i = 0; i < trees.size(); i++){
			root = trees.get(i);
			binaryTrees = new ArrayList<BinaryTreeNode>();
			// Go through all children of each tree (probably just one child in each)
			for (int j = 0; j < root.getChildren().size(); j++){
				// Make the tree binary, creating new nodes for existing nodes with more than 2 children
				binaryTree = convertTreeToBinary((SwcTreeNode)root.getChildren().get(j));

				// Calculate trajectories for path length adjustment in DiademMetric
				if (trajectoryThresholds != null){
					calculateTrajectories(binaryTree, trajectoryThresholds, zInPathDist);
				}
				// Remove nodes with only 1 child, and calculate path values (path length, surface, etc)
				binaryTree = removeContinuations(root, binaryTree, true, zInPathDist);

				// Put new binary tree in the list
				binaryTrees.add(binaryTree);
			}
			// Give root its children, now in binary tree form
			root.setChildren(binaryTrees);
		}
		
		//System.out.println("Done assembling trees from swc");
		
		return trees;
	}

	private static void addNode(
			SwcTreeNode node,
			SwcTreeNode parent,
			Map<Integer,SwcTreeNode> nodeMap,
			Map<SwcTreeNode,Integer> addNodeSideMap){

		if (node != null){
			//System.out.println("Parent: "+parent);
    		// Put new node in map
    		nodeMap.put(node.getNodeId(), node);
    		// Set node in meta data map (used for nodes with > 2 children)
    		addNodeSideMap.put(node, 0);
    		// If parent is null, then this is just the start of a new tree
    		// If parent is of a different node type, DO NOT ATTACH
    		if (parent != null && 
    				parent.getType() == node.getType()){
    			parent.addChildNode(node);
    		}
    	}
	}

	private static BinaryTreeNode removeContinuations(SwcTreeNode treeRoot, BinaryTreeNode binaryRoot, 
			boolean calcPathDist, boolean zInPathDist){
    	LinkedStack stack = new LinkedStack();
    	BinaryTreeNode node, headNode = binaryRoot, child;
    	SwcDataNode data, childData;

    	if (calcPathDist){
    		data = binaryRoot.getSwcData();
    		if (data.getSecondaryData() == null){
    			data.setSecondaryData(new SwcSecondaryData());
    		}
	    	if (zInPathDist){
	    		data.getSecondaryData().setPathLength(
	    			SwcDataUtils.getDistance(
	    					treeRoot.getSwcData(), data));
	    		data.getSecondaryData().setXYPathLength(SwcDataUtils.getXYDistance(
	    				treeRoot.getSwcData(), data));
	    	}
	    	else{
	    		data.getSecondaryData().setPathLength(
		    			SwcDataUtils.getXYDistance(
		    					treeRoot.getSwcData(), data));
	    		data.getSecondaryData().setXYPathLength(
	    				data.getSecondaryData().getPathLength());
	    	}
	    	data.getSecondaryData().setZPathLength(
	    			Math.abs(treeRoot.getSwcData().getZ() - binaryRoot.getSwcData().getZ()));
    	}

    	stack.push(binaryRoot);
    	while (!stack.isEmpty()){
    		node = (BinaryTreeNode)stack.pop();
    		data = node.getSwcData();
    		if (debug) System.out.println("RemoveCont Coords: "+data.getX()+","+data.getY());
    		
    		if (!node.isLeaf()){
    			stack.push(node.getLeft());
    			if (calcPathDist) calculatePathData(node.getLeft(), zInPathDist);
    			
				if (node.getRight() == null){
					// Accumulate node data in child since node will be removed
					child = node.getLeft();
					childData = child.getSwcData();
    				if (calcPathDist && data.getSecondaryData() != null){
    					childData.getSecondaryData().addData(data.getSecondaryData());
    				}
    				
    				if (node == headNode){
    	    			//System.out.println("Node has no parent");
    					// make solo child head if parent was head
    					headNode = child;
    				}
    				else{
						// add child to parent, remove self
	    				if (node.isLeft()){
	    					node.getParent().setLeft(child);
	    				}
	    				else{
	    					node.getParent().setRight(child);
	    				}
    				}
				}
				else{
					stack.push(node.getRight());
					if (calcPathDist) calculatePathData(node.getRight(), zInPathDist);
				}
    		}
    	}
    	return headNode;
	}
	
	private static void calculatePathData(BinaryTreeNode node, boolean zInPathDist){
		if (debug) System.out.println("Child: "+node.getSwcData());
		SwcDataNode data = node.getSwcData(), parentData = node.getParent().getSwcData();
		// Add child node to stack
		if (data.getSecondaryData() == null){
			data.setSecondaryData(new SwcSecondaryData());
		}
		// Calculate path distance and related metrics
		if (data.getSecondaryData().getPathLength() == 0){
	    	if (zInPathDist){
	    		data.getSecondaryData().setPathLength(
						SwcDataUtils.getDistance(parentData, data));
	    		data.getSecondaryData().setXYPathLength(
		    			SwcDataUtils.getXYDistance(parentData, data));
	    	}
	    	else{
	    		data.getSecondaryData().setPathLength(
						SwcDataUtils.getXYDistance(parentData, data));
	    		data.getSecondaryData().setXYPathLength(data.getSecondaryData().getPathLength());
	    	}
	    	data.getSecondaryData().setZPathLength(
	    			Math.abs(parentData.getZ() - data.getZ()));    				    	
			
	    	data.getSecondaryData().setSurfaceArea(
	    			data.getSecondaryData().getPathLength()
					* Math.PI * data.getDiameter());
	    	data.getSecondaryData().setVolume(
	    			data.getSecondaryData().getPathLength()
					* Math.PI * data.getRadiusSquared());
		}
	}
	
	/** 
	 * 
	 * @param root
	 * @param zInPathDist
	 */
	private static void calculateTrajectories(BinaryTreeNode root, SwcDataNode thresholds, boolean zInPathDist){
    	LinkedStack stack = new LinkedStack();
    	BinaryTreeNode node;
    	SwcDataNode data;

		if (root.getSwcData().getSecondaryData() == null){
			root.getSwcData().setSecondaryData(new SwcSecondaryData());
		}

		// Add first bifurcation
		node = root;
		while (node.getLeft() != null && node.getRight() == null){
			node = node.getLeft();
		}
		stack.push(node);

    	while (!stack.isEmpty()){
    		node = (BinaryTreeNode)stack.pop();
    		if (debug) System.out.println("Calculating trajectories for Node: "+node);
    		data = node.getSwcData();

    		if (data.getSecondaryData() == null){
    			data.setSecondaryData(new SwcSecondaryData());
    		}
    		
    		// Calculate for parent direction
    		if (!node.isRoot()){
    			data.getSecondaryData().setParentTrajectoryPoint(findParentTrajectory(node, thresholds));
    		}
    		
    		// If bifurcation, calculate trajectories down left and right children
    		if (!node.isLeaf()){
				// calculate for left
	    		data.getSecondaryData().setLeftTrajectoryPoint(
	    				findChildTrajectory(data, node.getLeft(), thresholds));
				
				// calculate for right
	    		data.getSecondaryData().setRightTrajectoryPoint(
	    				findChildTrajectory(data, node.getRight(), thresholds));

				// Add child bifurcations to stack
	    		addChildBifurcationsToStack(stack, node);
			}
    	}
	}
	
	private static void addChildBifurcationsToStack(LinkedStack stack, BinaryTreeNode node){
		BinaryTreeNode c = node.getLeft();
		// Traverse continuations until a bifurcation or termination (leaf) is found
		while(c.getRight() == null && c.getLeft() != null) {
			c = c.getLeft();
		} 
		stack.push(c);

		c = node.getRight();
		while(c.getRight() == null && c.getLeft() != null) {
			c = c.getLeft();
		}
		stack.push(c);
	}
	
	public static EuclideanPoint findParentTrajectory(
			TreeNode node, SwcDataNode thresholds){
		EuclideanPoint point = new EuclideanPoint();
		boolean doneX = false, doneZ = false;
		TreeNode c = node;
		SwcDataNode data = (SwcDataNode)node.getData(), cData = (SwcDataNode)c.getData(), prevData;
		
    	EuclideanPoint tmpPoint;
    	double xyDistance, zDistance;
    	
		while (!doneX || !doneZ) {
    		c = c.getTreeNodeParent();
    		prevData = cData;
    		cData = (SwcDataNode)c.getData();
    		if (!doneX){
        		// Calculate distance from bifurcation node to current continuation node 
        		xyDistance = SwcDataUtils.getXYDistance(data, cData);
	    		// If the continuation node is outside the threshold
	    		if (xyDistance > thresholds.getX()){
	    			tmpPoint = calculateTrajectoryXY(data, prevData, cData, thresholds.getX());
	    			point.setX(tmpPoint.getX());
	    			point.setY(tmpPoint.getY());
					doneX = true;
				}
	    		// Stop if root node is reached
	    		else if (c.getTreeNodeParent() == null){
					point.setX(cData.getX());
					point.setY(cData.getY());
					doneX = true;
				}
    		}
    		if (!doneZ){
        		zDistance = Math.abs(data.getZ() - cData.getZ());
        		if (zDistance > thresholds.getZ()){
        			point.setZ(calculateTrajectoryZ(data, prevData, cData, thresholds.getZ()));
        			doneZ = true;
        		}
        		// Stop if root node is reached
		    		else if (c.getTreeNodeParent() == null){
						point.setZ(cData.getZ());
						doneZ = true;
		    		}
    		}
		}
		return point;
	}
	
	public static EuclideanPoint findChildTrajectory(SwcDataNode data, BinaryTreeNode child, SwcDataNode thresholds){
		EuclideanPoint point = new EuclideanPoint(), tmpPoint;
		SwcDataNode cData, prevData = data;
		// calculate for parent
		double xyDistance, zDistance;
		boolean doneX = false, doneZ = false;

		while (!doneX || !doneZ) {
    		cData = child.getSwcData();
    		
    		if (!doneX){
        		xyDistance = SwcDataUtils.getXYDistance(data, cData);
	    		// If the continuation node is outside the threshold region 
	    		if (xyDistance > thresholds.getX()){
	    			tmpPoint = calculateTrajectoryXY(data, prevData, cData, thresholds.getX());
	    			point.setX(tmpPoint.getX());
	    			point.setY(tmpPoint.getY());
					doneX = true;
				}
	    		else if (child.isLeaf()){
	    			point.setX(cData.getX());
					point.setY(cData.getY());
					doneX = true;
	    		}
	    		else if (child.getRight() != null){
					point.setX(TRAJECTORY_NONE);
					point.setY(TRAJECTORY_NONE);
					doneX = true;
				}
    		}
    		
    		if (!doneZ){
        		zDistance = Math.abs(data.getZ() - cData.getZ());
	    		if (zDistance > thresholds.getZ()){
	    			point.setZ(calculateTrajectoryZ(data, prevData, cData, thresholds.getZ()));
	    			doneZ = true;
	    		}
	    		// If we've reached a termination then that is the trajectory node
	    		else if (child.isLeaf()){
					point.setZ(cData.getZ());
					doneZ = true;
				}
	    		// If we've reached a bifurcation, will rely on correct path's child for trajectory
				else if (child.getRight() != null){
					point.setZ(TRAJECTORY_NONE);
					doneZ = true;
				}
    		}
    		
    		// Traverse down
    		prevData = cData;
    		if (child.hasChildren()){
    			child = child.getLeft();
    		}
		}
		
		return point;
	}
	
	/**
	 * 
	 * @param origin
	 * @param first
	 * @param second
	 * @param threshold
	 * @return The point between the first and second point that is at a distance of "threshold" from the origin.
	 * This assumes that the first point is within threshold distance, and the second is beyond that distance. 
	 */
	public static EuclideanPoint calculateTrajectoryXY(
			EuclideanPoint origin, EuclideanPoint first, EuclideanPoint second, double threshold){
		EuclideanPoint point = new EuclideanPoint();
		// Initializing to maximum proportion
		double toFirst, toSecond, proportionAlongLine;
		
		toFirst = SwcDataUtils.getXYDistance(origin, first); 
		toSecond = SwcDataUtils.getXYDistance(origin, second);
		
		// Calculate the proportion along the segment to the point that is at the threshold distance from the origin node
		proportionAlongLine = (threshold - toFirst) / (toSecond - toFirst);
		
		// Set the coordinates based on the first point plus the distance to the threshold point
		point.setX(first.getX() + proportionAlongLine*(second.getX()-first.getX()));
		point.setY(first.getY() + proportionAlongLine*(second.getY()-first.getY()));
		return point;
	}
	public static double calculateTrajectoryZ(
			EuclideanPoint origin, EuclideanPoint first, EuclideanPoint second, double threshold){
		
		// Initializing to maximum proportion
		double toFirst, toSecond, proportionAlongLine;
		
		toFirst = Math.abs(origin.getZ() - first.getZ()); 
		toSecond = Math.abs(origin.getZ() - second.getZ());
		
		// Calculate the proportion along the segment to the point that is at the threshold distance from the origin node
		proportionAlongLine = (threshold - toFirst) / (toSecond - toFirst);

		// Set the coordinates based on the first point plus the distance to the threshold point
		return (first.getZ() + proportionAlongLine*(second.getZ()-first.getZ()));
	}
	
	// NOT assuming continuations have been removed
	@SuppressWarnings("unchecked")
	private static BinaryTreeNode convertTreeToBinary(SwcTreeNode head) throws Exception{
		if (head == null){
			return null;
		}
		BinaryTreeNode root = new BinaryTreeNode(), node, tmp;
		SwcTreeNode treeNode;
		root.setData(head);
		LinkedStack stack = new LinkedStack(), recentTreeStack = new LinkedStack();
		stack.push(root);
		if (debug) System.out.println("Converting Tree to Binary");
		while (!stack.isEmpty()){
			node = (BinaryTreeNode)stack.pop();
			treeNode = (SwcTreeNode)node.getData();
			if (debug) System.out.println("Node: "+treeNode.getData());
			// Copy tree node data to binary tree node data
			node.setData(treeNode.getSwcData());
			if (debug) System.out.println("Node data: "+node.getData());
			if (treeNode.hasChildren()){
				if (treeNode.getChildren().size() > 1){
					tmp = convertTreeNodeToBinaryTreeNode(
							(SwcDataNode)node.getData(),treeNode.getChildren());
					node.setLeft(tmp.getLeft());
					node.setRight(tmp.getRight());
					// Find new bottom nodes
					recentTreeStack.push(tmp.getLeft());
					recentTreeStack.push(tmp.getRight());
					while (!recentTreeStack.isEmpty()){
						node = (BinaryTreeNode)recentTreeStack.pop();
						if (node.isLeaf()){
							stack.push(node);
						}
						else{
							recentTreeStack.push(node.getLeft());
							recentTreeStack.push(node.getRight());
						}
					}
				}
				else{
					node.setLeft(new BinaryTreeNode());
					node.getLeft().setData((SwcTreeNode)treeNode.getChildren().get(0));
					stack.push(node.getLeft());	
				}
			}
		}
		
		return root;
	}

	private static BinaryTreeNode convertTreeNodeToBinaryTreeNode(
			SwcDataNode data, List<TreeNode> childList) {
		BinaryTreeNode newNode = null;
		SwcDataNode swcData1, swcData2, newSwcDataNode;
		NodePair nodePair, bestNodePair = new NodePair();
		Object object, best1, best2;

		if (childList.size() == 1){
			return new BinaryTreeNode(childList.get(0));
		}
		
		// Keep going until there is only one node left on top, it will be parent to a binary tree
		while (childList.size() > 1){
			best1 = childList.get(0); best2 = childList.get(1);
			bestNodePair = new NodePair();
			bestNodePair.setDistanceSq(-1);
			// Loop through multiple children, find closest two nodes
			for (int i = 0; i < childList.size()-1; i++){
				// Get data, need to know node type for pulling out of list
				if (childList.get(i).getClass().equals(SwcTreeNode.class)){
					swcData1 = (SwcDataNode)((SwcTreeNode)childList.get(i)).getSwcData();	
				}
				else{
					swcData1 = (SwcDataNode)((BinaryTreeNode)childList.get(i)).getData();
				}
				// Loop through the rest, comparing current child to all others
				for (int j = i+1; j < childList.size(); j++){
					if (childList.get(j).getClass().equals(SwcTreeNode.class)){
						swcData2 = (SwcDataNode)((SwcTreeNode)childList.get(j)).getSwcData();	
					}
					else{
						swcData2 = (SwcDataNode)((BinaryTreeNode)childList.get(j)).getData();
					}
					// create a node pair for determining closest two nodes
					nodePair = new NodePair(swcData1,swcData2);
					if (bestNodePair.getDistanceSq() == -1 
							|| nodePair.getDistanceSq() < bestNodePair.getDistanceSq()){
						bestNodePair = nodePair;
						best1 = childList.get(i);
						best2 = childList.get(j);
					}
				}
			}
			
			// Insert new node with as parent of closest nodes
			newNode = new BinaryTreeNode();
			newSwcDataNode = new SwcDataNode();
			newNode.setData(newSwcDataNode);
			newSwcDataNode.setNodeId(data.getNodeId());
			newSwcDataNode.setX(data.getX());
			newSwcDataNode.setY(data.getY());
			newSwcDataNode.setZ(data.getZ());
			newSwcDataNode.setType(data.getType());
			// Nodes share parent data
			newSwcDataNode.setSecondaryData(data.getSecondaryData());

			// Make closest nodes children of the new node
			object = best1;
			childList.remove(best1);
			if (object.getClass().equals(SwcTreeNode.class)){
				newNode.setLeft(new BinaryTreeNode());
				newNode.getLeft().setData((SwcTreeNode)object);
			}
			else{
				newNode.setLeft((BinaryTreeNode)object);
			}
			
			object = best2;
			childList.remove(best2);
			if (object.getClass().equals(SwcTreeNode.class)){
				newNode.setRight(new BinaryTreeNode());
				newNode.getRight().setData((SwcTreeNode)object);
			}
			else{
				newNode.setRight((BinaryTreeNode)object);
			}
			childList.add(newNode);
		}
		return newNode;
	}
	
	/**
	 * 
	 * @param inputFile
	 * @param scaleZ
	 * @return Map keyed by BinaryTreeNode, value is the node's associated swc path point.
	 * Path points are represented by SwcTreeNodes with a parent and 0, 1, or 2 children.
	 * @throws Exception
	 */
	public static Map<Integer,SwcTreeNode> convertSwcToTreeNodeMap(
			File inputFile, double scaleZ) throws Exception{
	    
		List<SwcTreeNode> rootList = convertSwcToSwcTreeList(inputFile, scaleZ);
		
		// Find terminations and assemble them into a list
		Map<Integer,SwcTreeNode> nodeMap = new HashMap<Integer,SwcTreeNode>();
		LinkedStack stack = new LinkedStack();
		SwcTreeNode root, node;
		for (int i = 0; i < rootList.size(); i++){
			root = rootList.get(i);
			stack.push(root);
			while (!stack.isEmpty()){
				node = (SwcTreeNode)stack.pop();
				if (node.isLeaf() || node.getChildren().size() > 1){
					nodeMap.put(node.getSwcData().getNodeId(), node);
				}
				for (int j = 0; j < node.getChildren().size(); j++){
					stack.push(node.getChildren().get(j));
				}
			}
		}
		
		if (debug) System.out.println("Done assembling trees from swc");
		
		return nodeMap;
	}
	
}