package org.krasnow.cng.diadem;

import java.io.File;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.zip.DataFormatException;

import org.krasnow.cng.data.ReadSWC;
import org.krasnow.cng.domain.BinaryTreeNode;
import org.krasnow.cng.domain.EuclideanPoint;
import org.krasnow.cng.domain.LinkedStack;
import org.krasnow.cng.domain.SwcDataNode;
import org.krasnow.cng.domain.SwcSecondaryData;
import org.krasnow.cng.domain.SwcTreeNode;
import org.krasnow.cng.utils.BinaryTreeUtils;
import org.krasnow.cng.utils.SwcDataUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * 
 * Calculates metric based on nodes and their connection to parent.
 * When parent is not found, searches up for match constrained by path length.
 * When node match is not found, searches up recursively for parent match,
 * then down for child match to determine continuation. Also constrained by
 * path length via error rate thresholds.
 *
 */
 public class DiademMetric {

	public final static int MODE_TEST_ENVIRONMENT = 0;
	public final static int MODE_COMPETITION = 1;
	public final static int MODE_OPEN = 2;

	// Determines whether certain input parameters will be acceptable, and whether debugging is available
	private static int mode = MODE_OPEN;
	private static boolean isTestEnvironment(){
		return MODE_TEST_ENVIRONMENT == mode;
	}

	public final static String PARAM_MICRONS = "microns";
	public final static String PARAM_TEST_DATA = "test-data";
	public final static String PARAM_GOLD_STANDARD_DATA = "gold-standard";
	public final static String PARAM_DATASET = "dataset";
	public final static String PARAM_WEIGHT_SCHEME = "weight-scheme";
	public final static String PARAM_XY_THRESHOLD = "xy-threshold";
	public final static String PARAM_XY_Z_RATIO = "xy-z-ratio";
	public final static String PARAM_Z_THRESHOLD = "z-threshold";
	public final static String PARAM_XY_PATH_THRESHOLD = "xyPathThresh";
	public final static String PARAM_Z_PATH_THRESHOLD = "zPathThresh";
	public final static String PARAM_Z_PATH = "z-path";
	public final static String PARAM_REMOVE_SPURS = "remove-spurs";
	public final static String PARAM_COUNT_EXCESS_NODE = "excess-nodes";
	public final static String PARAM_ALIGN_TREES_BY_ROOT = "align-roots";
	public final static String PARAM_TERMINAL_THRESHOLD = "terminal-threshold";
	public final static String PARAM_TERMINAL_PATH_THRESHOLD = "terminal-path-threshold";

	public final static String PARAM_LIST_MISSES = "list-misses";
	public final static String PARAM_LIST_CONTINUATIONS = "list-continuations";
	public final static String PARAM_LIST_DISTANT_MATCHES = "list-distant-matches";

	public final static boolean DEFAULT_WEIGHTED = true;
	public final static double DEFAULT_XY_THRESHOLD = 1.2;
	public final static double DEFAULT_Z_THRESHOLD = 1;
	public final static double DEFAULT_XY_PATH_THRESHOLD = 0.08;
	public final static double DEFAULT_Z_PATH_THRESHOLD = 0.20;
	public final static boolean DEFAULT_Z_PATH = false;
	public final static double DEFAULT_REMOVE_SPURS = 1;
	public final static double DEFAULT_TERMINAL_THRESHOLD = 0;
	public final static double DEFAULT_TERMINAL_PATH_THRESHOLD = 0;

	public final static double DEFAULT_PATH_LENGTH_THRESHOLD = .05;

	public final static int DATASET_NONE = 0;
	public final static int DATASET_CCF = 1;
	public final static int DATASET_HC_CA3_INTERNEURON = 2;
	public final static int DATASET_NC_LAYER1_AXONS = 3;
	public final static int DATASET_NM_PROJECTION = 4;
	public final static int DATASET_OLFACTORY_PROJECTION = 5;
	public final static int DATASET_VISUAL_CORTICAL = 6;

	public final static int WEIGHT_DEGREE = 1;
	public final static int WEIGHT_UNIFORM = 2;
	public final static int WEIGHT_SQRT_DEGREE = 3;
	public final static int WEIGHT_DEGREE_HARMONIC_MEAN = 4;
	public final static int WEIGHT_PATH_LENGTH = 5;
	
	// Determines whether thresholds are micron or pixel based
	private boolean microns = false;
	private File testSwcFile;
	private File goldSwcFile;
	private int dataset = 0;
	private double XYThreshold = DEFAULT_XY_THRESHOLD;
	private double rawZThreshold = DEFAULT_Z_THRESHOLD;
	private double ZThreshold;
	// Multiplier of Z component to keep XY and Z components in the same units
	private double scaleZ = 1;
	private double XYPathErrorThreshold = DEFAULT_PATH_LENGTH_THRESHOLD;
	private double ZPathErrorThreshold = DEFAULT_PATH_LENGTH_THRESHOLD;
	private double terminalThreshold = DEFAULT_TERMINAL_THRESHOLD;
	private double terminalPathThreshold = DEFAULT_TERMINAL_PATH_THRESHOLD;
	private double LocalPathErrorThreshold = 0.4; // Approximate % error for hypotenuse to sides
	private int weightMode = WEIGHT_DEGREE;
	private boolean countExcessNodes = true;
	private boolean alignTreesByRoot = false;
	private boolean debug = false;
	private boolean listMisses = false;
	private boolean listContinuations = false;
	private boolean listDistantMatches = false;
	// Determines whether Z component is considered while calculating path length
	private boolean zPath = false;
	// Used to see debug details for node(s) at xy coordinates 
	private double x;
	private double y;
	SwcDataNode xyCheck = new SwcDataNode();
	// Threshold under which terminal nodes are removed
	private double removeSpurs = 0;
	// Set based on whether current/target node is which XY threshold of x and y parameters
	boolean writeDetails;
	private List<SwcDataNode> offsets = new ArrayList<SwcDataNode>();
	
	//private boolean calculated;
	
	private List<BinaryTreeNode> misses = new ArrayList<BinaryTreeNode>();
	private List<BinaryTreeNode> continuations = new ArrayList<BinaryTreeNode>();
	private List<BinaryTreeNode> distantMatches = new ArrayList<BinaryTreeNode>();
	private List<BinaryTreeNode> spurList = new ArrayList<BinaryTreeNode>();
	private Map<BinaryTreeNode,Double> weightMap = new HashMap<BinaryTreeNode,Double>();
	private Map<BinaryTreeNode,Number> excessNodes = new LinkedHashMap<BinaryTreeNode,Number>();

	private double scoreSum, weightSum, quantityScoreSum;

	private double directMatchScore;
	private double qualityScore;
	private double finalScore;

	// Specifically for Neuromuscular Projection Fibers (dataset 4)
	private final static double ROSETTE_THRESHOLD = 10;
	private Map<Integer,SwcTreeNode> testTreePathNodes = null;
	private Map<Integer,SwcTreeNode> goldTreePathNodes = null;

	// For ensuring nodes aren't used multiple times
	private Map<BinaryTreeNode,BinaryTreeNode> matches = new HashMap<BinaryTreeNode,BinaryTreeNode>();
	
	public DiademMetric(File testSwcFile, File goldSwcFile, int dataset){
		this.testSwcFile = testSwcFile;
		this.goldSwcFile = goldSwcFile;
		setDataset(dataset);
	}

	public DiademMetric(File testSwcFile, File goldSwcFile){
		this.testSwcFile = testSwcFile;
		this.goldSwcFile = goldSwcFile;
	}

	public void setDataset(int dataset) {
		this.dataset = dataset;
		double pixelPerMic = 1;
		
		switch(dataset){
		case DATASET_CCF:
			/* 26.67 pixels/micron, 0.0375 microns/pixel
			   3 images/micron, 0.33 microns/image */
			pixelPerMic = 26.67;
			// Units in pixels and images
			XYThreshold = 37.33;
			scaleZ = 8.8; // pixels/image
			ZThreshold = 4*scaleZ; // ZThresh in pixels
			
			XYPathErrorThreshold = 0.075;
			ZPathErrorThreshold = 0.18;
			zPath = true;
			// No spur removal
			removeSpurs = 0;
			break;
		case DATASET_HC_CA3_INTERNEURON:
			/* 4.6 pixels/micron, 0.217391 microns/pixel
			   3 images/micron, 0.33 microns/image */
			pixelPerMic = 4.6;
			
			XYThreshold = 11;
			scaleZ = 1.52; // pixels/image
			//ZThreshold = 3 * scaleZ; // ZThresh in pixels
			ZThreshold = 14 * scaleZ; // ZThresh in pixels
			XYPathErrorThreshold = 0.08;
			// Z path length error is not to be constrained
			ZPathErrorThreshold = 100;
			zPath = false;
			
			removeSpurs = 23;
			break;
		case DATASET_NC_LAYER1_AXONS:
			/* 3.4 pixels/micron, 0.2941 microns/pixel
			   1 images/micron, 1 microns/image */
			pixelPerMic = 3.4;

			XYThreshold = 4.76;
			scaleZ = 3.4; // pixels/image
			ZThreshold = 5 * scaleZ; // ZThresh in pixels
			XYPathErrorThreshold = 0.07;
			ZPathErrorThreshold = 0.18;
			
			removeSpurs = 17;
			break;
		case DATASET_NM_PROJECTION:
			/* 26.67 pixels/micron, 0.0375 microns/pixel
			   5 images/micron, 0.2 microns/image */
			pixelPerMic = 26.67;

			XYThreshold = 32;
			XYPathErrorThreshold = 0.04;
			// Z path error will be zero, threshold is irrelevant
			ZPathErrorThreshold = 100;
			scaleZ = 0;
			removeSpurs = 0;
			terminalThreshold = ROSETTE_THRESHOLD;
			break;
		case DATASET_OLFACTORY_PROJECTION:
			/* 3.034 pixels/micron, 0.3296485 microns/pixel
			   1 images/micron, 1 microns/image */
			pixelPerMic = 3.034;

			XYThreshold = 3.94;
			scaleZ = 3.034; // pixels/image
			ZThreshold = 5 * scaleZ; // ZThresh in pixels
			
			XYPathErrorThreshold = 0.08;
			ZPathErrorThreshold = 0.20;
			
			zPath = true;
			removeSpurs = 6;
			break;
		case DATASET_VISUAL_CORTICAL:
			/* ? pixels/micron, ? microns/pixel
			   ? images/micron, ? microns/image */
			//pixelPerMic = 1; // Currently not defined

			XYThreshold = 9;
			scaleZ = 2.93; // pixels/image
			ZThreshold = 6 * scaleZ; // ZThresh in pixels
			
			XYPathErrorThreshold = 0.08;
			ZPathErrorThreshold = 0.20;
			
			zPath = true;
			removeSpurs = 50;
			break;
		}
		if (microns){
			XYThreshold = XYThreshold / pixelPerMic;
			ZThreshold = ZThreshold / pixelPerMic;
			removeSpurs = removeSpurs / pixelPerMic;
		}
		if (isTestEnvironment()) System.out.println("XYThreshold: "+XYThreshold+"; ZThreshold: "+ZThreshold);
	}

	public int getDataset() {
		return dataset;
	}

	public boolean isDebug() {
		return debug;
	}

	public void setDebug(boolean debug) {
		this.debug = debug;
	}

	public boolean isAlignTreesByRoot() {
		return alignTreesByRoot;
	}

	public void setAlignTreesByRoot(boolean alignTreesByRoot) {
		this.alignTreesByRoot = alignTreesByRoot;
	}

	public boolean isListContinuations() {
		return listContinuations;
	}

	public void setListContinuations(boolean listContinuations) {
		this.listContinuations = listContinuations;
	}

	public boolean isListDistantMatches() {
		return listDistantMatches;
	}

	public void setListDistantMatches(boolean listDistantMatches) {
		this.listDistantMatches = listDistantMatches;
	}

	public boolean isListMisses() {
		return listMisses;
	}

	public void setListMisses(boolean listMisses) {
		this.listMisses = listMisses;
	}

	public File getGoldSwcFile() {
		return goldSwcFile;
	}

	public void setGoldSwcFile(File goldSwcFile) {
		this.goldSwcFile = goldSwcFile;
		//calculated = false;
	}

	public int getWeightMode() {
		return weightMode;
	}

	public void setWeightMode(int weightMode) {
		this.weightMode = weightMode;
	}

	public File getTestSwcFile() {
		return testSwcFile;
	}

	public void setTestSwcFile(File testSwcFile) {
		this.testSwcFile = testSwcFile;
	}

	public boolean isMicrons() {
		return microns;
	}

	public void setMicrons(boolean microns) {
		this.microns = microns;
	}

	public double getXYThreshold() {
		return XYThreshold;
	}

	public void setXYThreshold(double threshold) {
		XYThreshold = threshold;
	}

	public double getTerminalThreshold() {
		return terminalThreshold;
	}

	public void setTerminalThreshold(double terminalThreshold) {
		this.terminalThreshold = terminalThreshold;
	}

	public double getTerminalPathThreshold() {
		return terminalPathThreshold;
	}

	public void setTerminalPathThreshold(double terminalPathThreshold) {
		this.terminalPathThreshold = terminalPathThreshold;
	}

	public double getZThreshold() {
		return ZThreshold;
	}

	/**
	 * 
	 * @param threshold
	 * Sets the Z threshold in terms of pixels. This is the value used in the metric, but normally is based in 
	 * the XY/Z ratio and the raw Z threshold (in # of images)
	 */
	public void setZThreshold(double threshold) {
		ZThreshold = threshold;
	}

	public double getRawZThreshold() {
		return rawZThreshold;
	}

	/**
	 * 
	 * @param threshold
	 * Sets the raw Z threshold then calculates the ZThreshold using rawZThreshold and scaleZ (the XY/Z ratio)
	 */
	public void setRawZThreshold(double threshold) {
		rawZThreshold = threshold;
		ZThreshold = rawZThreshold * scaleZ;
	}

	public double getXYPathErrorThreshold() {
		return XYPathErrorThreshold;
	}

	public void setXYPathErrorThreshold(double pathLengthThreshold) {
		XYPathErrorThreshold = pathLengthThreshold;
	}

	public double getZPathErrorThreshold() {
		return ZPathErrorThreshold;
	}

	public boolean isZPath() {
		return zPath;
	}

	public void setZPath(boolean path) {
		zPath = path;
	}

	public double getRemoveSpurs() {
		return removeSpurs;
	}

	public void setRemoveSpurs(double removeSpurs) {
		this.removeSpurs = removeSpurs;
	}


	public boolean isCountExcessNodes() {
		return countExcessNodes;
	}

	public void setCountExcessNodes(boolean countExcessNodes) {
		this.countExcessNodes = countExcessNodes;
	}

	public void setZPathErrorThreshold(double pathLengthThreshold) {
		ZPathErrorThreshold = pathLengthThreshold;
	}

	public double getX() {
		return x;
	}

	public void setX(double x) {
		this.x = x;
	}

	public double getY() {
		return y;
	}

	public void setY(double y) {
		this.y = y;
	}

	public double getFinalScore(){
		return finalScore;
	}
	
	public double getScaleZ(){
		return scaleZ;
	}

	/**
	 * 
	 * @param scaleZ
	 * Sets the XY/Z ratio in member "scaleZ", then calculates the pixel 
	 * ZThreshold using the scaling factor and the rawZThreshold (in # of images)
	 */
	public void setScaleZ(double scaleZ){
		this.scaleZ = scaleZ;
		ZThreshold = rawZThreshold * scaleZ;
	}
	
	// These return statements copy lists and maps, providing information about how individual nodes were classified and scored
	
	/**
	 * 
	 * @return List of missed gold standard nodes
	 */
	public List<BinaryTreeNode> getMisses() {
		return (misses == null ? null : new ArrayList<BinaryTreeNode>(misses));
	}

	/**
	 * 
	 * @return List of gold standard nodes marked as continuations: no direct match to a test node
	 */
	public List<BinaryTreeNode> getContinuations() {
		return (continuations == null ? null : new ArrayList<BinaryTreeNode>(continuations));
	}

	/**
	 * 
	 * @return List of gold standard nodes marked as distant matches: the match was outside the threshold, but within a distance of 3 thresholds and with confirmed local ancestor and descendant connectivity
	 */
	public List<BinaryTreeNode> getDistantMatches() {
		return (distantMatches == null ? null : new ArrayList<BinaryTreeNode>(distantMatches));
	}

	/**
	 * 
	 * @return List of gold standard and test nodes marked as spurs
	 */
	public List<BinaryTreeNode> getSpurList() {
		return (spurList == null ? null : new ArrayList<BinaryTreeNode>(spurList));
	}

	/**
	 * 
	 * @return WeightMap keyed by node with Double values of the node's weight in scoring
	 */
	public Map<BinaryTreeNode,Double> getWeightMap() {
		return (weightMap == null ? null : new HashMap<BinaryTreeNode,Double>(weightMap));
	}

	/**
	 * 
	 * @return ExcessNodes keyed by test reconstruction nodes considered excess with Double values of each nodes' weight in scoring
	 */
	public Map<BinaryTreeNode,Number> getExcessNodes() {
		return (excessNodes == null ? null : new HashMap<BinaryTreeNode,Number>(excessNodes));
	}

	/**
	 * 
	 * @return The score produced by considering continuations as misses
	 */
	public double getDirectMatchScore() {
		return directMatchScore;
	}

	/**
	 * 
	 * @return Matches map keyed by all matched nodes with Double values each node's weight in scoring
	 */
	public Map<BinaryTreeNode,BinaryTreeNode> getMatches() {
		return matches;
	}

	public void scoreReconstruction() throws Exception{
		
		List<SwcTreeNode> testTreeRoots, goldTreeRoots;
		SwcTreeNode testTreeRoot, goldTreeRoot;
		List<File> goldFileList = new ArrayList<File>(), testFileList = new ArrayList<File>();

		// Gets all gold standard and test swc files in the directories
		if (goldSwcFile.isDirectory()){
			File[] fileList = goldSwcFile.listFiles();
			File goldFile, testFile;
			String filename;
			for (int i = 0; i < fileList.length; i++){
				goldFile = fileList[i];
				if (!goldFile.isDirectory()){
					filename = goldFile.getName();
					if (filename.endsWith(".swc") || filename.endsWith(".SWC")){
						testFile = new File(testSwcFile.getAbsolutePath()+"/"+filename);
						if (testFile.exists()){
							goldFileList.add(goldFile);
							testFileList.add(testFile);
						}
					}
				}
			}
		}
		else{
			// For case in which parameter values are swc files rather than directories
			goldFileList.add(goldSwcFile);
			testFileList.add(testSwcFile);
		}

		scoreSum = quantityScoreSum = weightSum = 0;

		// Loop through and read in all swc files
		for (int i = 0; i < goldFileList.size(); i++){
			goldSwcFile = goldFileList.get(i);
			testSwcFile = testFileList.get(i);

			SwcDataNode thresholds = new SwcDataNode();
			thresholds.setX(XYThreshold);
			thresholds.setZ(ZThreshold);
			goldTreeRoots = ReadSWC.convertSwcToBinaryTreeList(goldSwcFile, zPath, scaleZ, thresholds);
			goldTreeRoot = (SwcTreeNode)goldTreeRoots.get(0);
			if (goldTreeRoots.size() > 1){
				System.out.println("Warning: Only the first tree in the gold standard file is being used. There are "+goldTreeRoots.size()+" trees total.");
			}
			testTreeRoots = ReadSWC.convertSwcToBinaryTreeList(testSwcFile, zPath, scaleZ, thresholds);
			testTreeRoot = testTreeRoots.get(0);
			if (testTreeRoots.size() > 1){
				System.out.println("Warning: Only the first tree in the test file is being used. There are "+testTreeRoots.size()+" trees total.");
			}
			
			if (terminalThreshold > 0){
				// Necessary for handling terminations within rosettes
				testTreePathNodes = ReadSWC.convertSwcToTreeNodeMap(testSwcFile, scaleZ);
				goldTreePathNodes = ReadSWC.convertSwcToTreeNodeMap(goldSwcFile, scaleZ);
			}
			// target nodes near xyCheck will display detailed information (only in test environment)
			xyCheck.setX(getX());
			xyCheck.setY(getY());
			
			if (removeSpurs > 0){
				// Only remove spurs of gold standard. Remove test spurs prior to determining excess
				removeSpurs(goldTreeRoot, removeSpurs);
			}
			generateNodeWeights(goldTreeRoot);
			
			// This is where the real business is done
			scoreTrees(testTreeRoot, goldTreeRoot);
		}

		// Present output
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(3);
		System.out.println("Score: "+nf.format(finalScore));

		// List misses
		BinaryTreeNode node;
		if (listMisses){
			System.out.println();
			if (misses.size() > 0){
				System.out.println("Nodes that were missed (position and weight):");
				for (int i = 0; i < misses.size(); i++){
					node = misses.get(i);
					System.out.println(getPositionString(node)+" "+weightMap.get(node));
				}
			}
			else{
				System.out.println("Nodes that were missed: none");
			}

			System.out.println();
			if (excessNodes.size() > 0){
				System.out.println("Extra nodes in test reconstruction (position and weight):");
				for (BinaryTreeNode excessNode : excessNodes.keySet()){
					System.out.println(getPositionString(excessNode)+" "+excessNodes.get(excessNode));
				}
			}
			else{
				System.out.println("Extra nodes in test reconstruction: none");
			}
		}

		// Present output
		if (isTestEnvironment()){
			System.out.println("Direct hit score: "+nf.format(directMatchScore));
			switch (weightMode){
			case WEIGHT_DEGREE:
				System.out.println("Weighted by degree");
				break;
			case WEIGHT_SQRT_DEGREE:
				System.out.println("Weighted by sqrt(degree)");
				break;
			case WEIGHT_DEGREE_HARMONIC_MEAN:
				System.out.println("Weighted by the harmonic mean of subtree degree");
				break;
			case WEIGHT_UNIFORM:
				default:
					System.out.println("Weighted uniformly");
			}
			System.out.println("Score including continuations: "+nf.format(qualityScore));
			System.out.println("Total adjusted for excess nodes: "+nf.format(finalScore));
	
			// List distant matches
			if (listDistantMatches){
				System.out.println();
				if (distantMatches.size() > 0){
					System.out.println("Distant Matches:");
					for (int i = 0; i < distantMatches.size(); i++){
						node = distantMatches.get(i);
						System.out.println(getPositionString(node)+" "+node.getDegree());
					}
				}
				else{
					System.out.println("Distant Matches: none");
				}
			}
			// List continuations
			if (listContinuations){
				System.out.println();
				if (continuations.size() > 0){
					System.out.println("Continuation nodes (position and weight):");
					for (int i = 0; i < continuations.size(); i++){
						node = continuations.get(i);
						System.out.println(getPositionString(node)+" "+weightMap.get(node));
					}
				}
				else{
					System.out.println("Continuation nodes: none");
				}
			}
		}
	}

	/**
	 * 
	 * @param testRoot
	 * @param goldRoot
	 * @return
	 */
	public void scoreTrees(
			SwcTreeNode testRoot, SwcTreeNode goldRoot) throws Exception{
		BinaryTreeNode match;

		if (alignTreesByRoot){
			alignRoots(testRoot,goldRoot);
		}
		
		// Weights
		double weight;
		List<BinaryTreeNode> testTrees = testRoot.getChildren();
		List<BinaryTreeNode> goldTrees = goldRoot.getChildren();
		int numberOfNodes = 0;
		
		SwcDataNode goldData;

		// For roots, create nodes and set as match
		goldRoot.getSwcData().setSecondaryData(new SwcSecondaryData());
		// Root is coregistered and so will have no trajectory calculation (trajectory is self for simplicity)
		goldRoot.getSwcData().getSecondaryData().setLeftTrajectoryPoint(goldRoot.getSwcData());
		goldRoot.getSwcData().getSecondaryData().setRightTrajectoryPoint(goldRoot.getSwcData());
		goldRoot.getSwcData().getSecondaryData().setPathLength(0);
		BinaryTreeNode trueBaselineRoot = new BinaryTreeNode(goldRoot.getSwcData());
		testRoot.getSwcData().setSecondaryData(new SwcSecondaryData());
		testRoot.getSwcData().getSecondaryData().setLeftTrajectoryPoint(testRoot.getSwcData());
		testRoot.getSwcData().getSecondaryData().setRightTrajectoryPoint(testRoot.getSwcData());
		testRoot.getSwcData().getSecondaryData().setPathLength(0);
		BinaryTreeNode trueSampleRoot = new BinaryTreeNode(testRoot.getSwcData());
		matches.put(trueBaselineRoot, trueSampleRoot);
		matches.put(trueSampleRoot, trueBaselineRoot);
		
		// Traverses test tree and returns a list of nodes
		List<BinaryTreeNode> testNodeList = new ArrayList<BinaryTreeNode>();
		BinaryTreeNode rootNode;
		for (int i = 0; i < testTrees.size(); i++){
			rootNode = testTrees.get(i);
			testNodeList.addAll(
					BinaryTreeUtils.createNodeList(rootNode));
			// Give first node a stable parent, the root of all trees should be a given
			rootNode.setParent(trueSampleRoot, false);
		}

		// Traverses gold tree and returns a list of nodes
		List<BinaryTreeNode> goldNodeList = new ArrayList<BinaryTreeNode>();
		for (int i = 0; i < goldTrees.size(); i++){
			rootNode = goldTrees.get(i);
			goldNodeList.addAll(
					BinaryTreeUtils.createNodeList(rootNode));
			// Give first node a stable parent, the root of all trees should be a given
			rootNode.setParent(trueBaselineRoot, false);
		}

		// Stack used to keep track of nodes to be scored
		LinkedStack stack = new LinkedStack();
		BinaryTreeNode goldNode;
		
		numberOfNodes -= spurList.size();
		
		// Score nodes in each gold standard tree
		for (int i = 0; i < goldTrees.size(); i++){
			goldNode = goldTrees.get(i);
			
			numberOfNodes += goldNode.getTreeSize();
			
			// Score each node (including "root", which is the first bifurcation)
			stack.push(goldNode);
			
			// Traverse all of the nodes in the current gold standard tree
			while (!stack.isEmpty()){
				goldNode = (BinaryTreeNode)stack.pop();
				goldData = (SwcDataNode)goldNode.getData();
				
				// Add this node's children to the stack
				if (goldNode.hasChildren()){
					stack.push(goldNode.getLeft());
					stack.push(goldNode.getRight());
				}
				
				// Determine whether to write details
				writeDetails = false;
				if (isTestEnvironment() && isWithinXYThreshold(xyCheck, goldData)){
					writeDetails = true;
				}

				// Only score node if it isn't a spur or single spur parent
				if (spurList.contains(goldNode)){
					if (writeDetails) System.out.println("\nBaselineNode: "+getPositionString(goldNode)+ " is a spur");
				}
				else{
					// Set weight
					weight = ((Double)weightMap.get(goldNode)).doubleValue();
					weightSum += weight;
		
					if (writeDetails || debug){
						System.out.println("\nBaselineNode: "+getPositionString(goldNode));
						System.out.println("Degree: "+goldNode.getDegree());
					}

					// Find match for target node (search for potentials, confirmation, and selection of best)
					match = getClosestMatch(goldNode, goldNodeList, testNodeList);
					if (match == null && goldNode.isLeaf() && terminalThreshold > 0){
						// Use the extended terminal search method
						if (dataset == DATASET_NM_PROJECTION){
							match = getNMPTerminationMatch(goldNode, goldNodeList, testNodeList);
						}
						else{
							match = getExtendedTerminationMatch(goldNode, goldNodeList, testNodeList);
						}
					}
					
					// Update quantity score based on whether a match has been found
					if (match != null){
						// Update node to save time on potential parent search
						matches.put(goldNode, match);
						// Update node uses to prevent node reuse
						matches.put(match, goldNode);
						// increment score sum
						scoreSum += weight;
		
						// Increment quantity (non-continuations) 
						quantityScoreSum++;
						if (writeDetails) System.out.println("Scoring direct match to "+getPositionString(match));
					}
					else{
						if (writeDetails) System.out.println("No direct match.");
						// Add to list of misses
						misses.add(goldNode);
					}
				}
			}
		}

		// Loop through bifurcation misses to determine if any might be a continuation
		if (isTestEnvironment()) System.out.println("\nLooking for continuations");
		for (int i = 0; i < misses.size(); i++){
			goldNode = misses.get(i);
			goldData = goldNode.getSwcData();
			writeDetails = false;
			if (isWithinXYThreshold(xyCheck, goldData)){
				writeDetails = true;
			}

			if (!goldNode.isLeaf()){
				if (writeDetails || debug) System.out.println("\nTarget "+getPositionString(goldNode));

				// If a continuation match via (grand)parent and (grand)child was found
				if (isContinuation(goldNode, testNodeList)){
					// increment score sum
					weight = weightMap.get(goldNode);
					scoreSum += weight;
					if (writeDetails || debug) System.out.println("Found continuation");
					// This node is now a continuation, not a miss
					misses.remove(i);
					i--;
				}
			}
		}
		
		// Calculate final scores
		if (weightSum > 0){
			directMatchScore = quantityScoreSum / numberOfNodes;

			// Quality score is average weighted node quality, without accounting for excess nodes
			qualityScore = scoreSum / weightSum;
			
			if (isTestEnvironment()) {
				System.out.println("WeightSum: "+weightSum);
				System.out.println("ScoreSum: "+scoreSum);
			}
			
			// Remove spurs prior to determining excess
			if (removeSpurs > 0){
				removeSpurs(testRoot, removeSpurs);
			}
			// Excess nodes weighed by degree of excess terms without any matches in between
			if (countExcessNodes){
				weightSum += weighExcess(testRoot, goldTrees);
			}
			finalScore = scoreSum / weightSum;
		}
		
	}
	
	/**
	 * 
	 * @param goldNode
	 * @param testNodeList
	 * @return BinaryTreeNode
	 *  Returns the best match to the target (gold) node
	 */
	private BinaryTreeNode getClosestMatch(
			BinaryTreeNode goldNode, List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		// Find nearest node to parent
		List<BinaryTreeNode> nearbyNodesList = findNearestNodes(goldNode, testNodeList);
		BinaryTreeNode testNode, bestMatch = null;
		
		if (writeDetails) System.out.println(nearbyNodesList.size()+" nearby");
		// Check whether nearest node(s) has parent that matches one of gold node's parents
		List<BinaryTreeNode> matchList = new ArrayList<BinaryTreeNode>();
		for (int j = 0; j < nearbyNodesList.size(); j++){
			testNode = nearbyNodesList.get(j);
			// Check whether nodes' parents match (based on position and path length)
			if (writeDetails) System.out.println("Checking path lengths for match: "+getPositionString(testNode)+"; Dist: "+getDistance(goldNode, testNode));
			if (nodeMatches(goldNode, testNode, goldNodeList, testNodeList)){
				matchList.add(testNode);
			}
		}
		
		if (matchList.size() == 1){
			bestMatch = matchList.get(0);
		}
		else if (matchList.size() > 1){
			// Find the best match of those found based on descendant connectivity and/or proximity
			bestMatch = determineBestMatch(matchList, goldNode, testNodeList);
		}
		if (bestMatch != null && writeDetails) 
			System.out.println("Best match is "+getPositionString(bestMatch));

		return bestMatch;
	}

	/**
	 * 
	 * @param matchList
	 * @param goldNode
	 * @param testNodeList
	 * @return best BinaryTreeNode based on descendant connectivity and/or proximity
	 */
	private BinaryTreeNode determineBestMatch(List<BinaryTreeNode> matchList, BinaryTreeNode goldNode,
			List<BinaryTreeNode> testNodeList){

		if (writeDetails) System.out.println("Determining best match");
		List<BinaryTreeNode> confirmList = new ArrayList<BinaryTreeNode>();
		BinaryTreeNode descendantNode, testDescendantNode, testNode, match;
		
		// For correction of test path length
		SwcDataNode goldData = goldNode.getSwcData();
		EuclideanPoint targetTrajectory, descendantTrajectory;
		Map<BinaryTreeNode,Boolean> descendantBranchLeft = new HashMap<BinaryTreeNode,Boolean>();
		Map<BinaryTreeNode,EuclideanPoint> trajectoryMap = new HashMap<BinaryTreeNode,EuclideanPoint>();
		
		if (goldNode.hasChildren()){
			List<BinaryTreeNode> nearbyList;
			
			// Path distances
			Map<BinaryTreeNode,SwcSecondaryData> pathLengthMap = new HashMap<BinaryTreeNode,SwcSecondaryData>();
			pathLengthMap.put(goldNode, new SwcSecondaryData());
			SwcSecondaryData parentPathLength, goldPathLength, testPathLength;
			double testXYPathLength, testZPathLength;
			
			// Euclidean distances
			List<SwcSecondaryData> targetDistancesList = new ArrayList<SwcSecondaryData>();
			SwcSecondaryData descendantDistances, targetDistances;
			for (int i = 0; i < matchList.size(); i++){
				testNode = matchList.get(i);
				targetDistances = new SwcSecondaryData();
				targetDistances.setXYPathLength(getXYDistance(goldNode, testNode));
				targetDistances.setZPathLength(getZDistance(goldNode, testNode));
				targetDistancesList.add(targetDistances);
			}
			
			// Traversing gold node descendants
			List<BinaryTreeNode> currentList = new ArrayList<BinaryTreeNode>(), nextList;
			currentList.add(goldNode.getLeft());
			trajectoryMap.put(goldNode.getLeft(), goldData.getSecondaryData().getLeftTrajectoryPoint());
			descendantBranchLeft.put(goldNode.getLeft(), true);
			currentList.add(goldNode.getRight());
			trajectoryMap.put(goldNode.getRight(), goldData.getSecondaryData().getRightTrajectoryPoint());
			descendantBranchLeft.put(goldNode.getRight(), false);
			
			while (currentList.size() > 0 && confirmList.size() == 0){
				nextList = new ArrayList<BinaryTreeNode>();

				// Loop through nodes at current branch order
				for (int i = 0; i < currentList.size(); i++){
					descendantNode = currentList.get(i);
					targetTrajectory = trajectoryMap.get(descendantNode);
					// Must determine trajectory based on specific path
					if (targetTrajectory.getX() == ReadSWC.TRAJECTORY_NONE || targetTrajectory.getZ() == ReadSWC.TRAJECTORY_NONE){
						EuclideanPoint tmpTrajectory = getTrajectoryForPath(goldNode, descendantNode);
						if (targetTrajectory.getX() == ReadSWC.TRAJECTORY_NONE){
							targetTrajectory.setX(tmpTrajectory.getX());
							targetTrajectory.setY(tmpTrajectory.getY());
						}
						if (targetTrajectory.getZ() == ReadSWC.TRAJECTORY_NONE){
							targetTrajectory.setZ(tmpTrajectory.getZ());
						}
					}

					descendantTrajectory = descendantNode.getSwcData().getSecondaryData().getParentTrajectoryPoint();

					// Add children to next list
					if (descendantNode.hasChildren()){
						nextList.add(descendantNode.getLeft());
						nextList.add(descendantNode.getRight());
						trajectoryMap.put(descendantNode.getLeft(), targetTrajectory);
						trajectoryMap.put(descendantNode.getRight(), targetTrajectory);
					}

					if (writeDetails) System.out.println("Descendant node: "+descendantNode.getSwcData().getPositionString());

					// Calculate path length from target node to current descendant
					goldPathLength = new SwcSecondaryData();
					goldPathLength.incrementPathLengths(descendantNode.getSwcData().getSecondaryData());
					parentPathLength = (SwcSecondaryData)pathLengthMap.get(descendantNode.getParent());
					goldPathLength.incrementPathLengths(parentPathLength);
					pathLengthMap.put(descendantNode, goldPathLength);
					
					// Find test nodes that match current descendant
					nearbyList = findNearestNodes(descendantNode, testNodeList, true);
					
					// Determine if any of these nodes is a descendant of a target node match
					for (int j = 0; j < nearbyList.size(); j++){
						// testDescendant stays set as this, testNode climbs to look for match node
						testDescendantNode = testNode = nearbyList.get(j);
						testPathLength = new SwcSecondaryData();
		
						if (writeDetails) System.out.println("Nearby descendant test node: "+testNode.getSwcData().getPositionString());

						// Descendant euclidean distances
						descendantDistances = new SwcSecondaryData();
						descendantDistances.setXYPathLength(getXYDistance(descendantNode, testNode));
						descendantDistances.setZPathLength(getZDistance(descendantNode, testNode));
		
						// Climb matched descendant until a target match is found or the root is hit
						while (testNode.getParent() != null){
							testPathLength.incrementPathLengths(testNode.getSwcData().getSecondaryData());
		
							testNode = testNode.getParent();
							// Check whether one of the match nodes is encountered while climbing
							for (int k = 0; k < matchList.size(); k++){
								match = matchList.get(k);
								if (testNode == match){
									if (writeDetails) System.out.println("Match (descendant's ancestor): "+getPositionString(match));
									targetDistances = targetDistancesList.get(k);
									// Found a target match, need to check path length
									
									// Calculate test XY and Z path length modified by end node positions
									testXYPathLength = testPathLength.getXYPathLength() 
										+ getEndNodeXYDistanceDifference(goldData, targetTrajectory, testNode.getSwcData())
										+ getEndNodeXYDistanceDifference(descendantNode.getSwcData(), descendantTrajectory, testDescendantNode.getSwcData());
									testZPathLength = testPathLength.getZPathLength() 
										+ getEndNodeZDistanceDifference(goldData, targetTrajectory, testNode.getSwcData())
										+ getEndNodeZDistanceDifference(descendantNode.getSwcData(), descendantTrajectory, testDescendantNode.getSwcData());

									// If path lengths are within threshold, return this specific target match
									if (pathLengthMatches(goldPathLength, testXYPathLength, testZPathLength) < 1){
										confirmList.add(match);
										matchList.remove(k);
										k--;
									}
								}
							}
						}
					}
				}
				currentList = nextList;
			}
		}
		
		// Only one confirmed, so return it
		if (confirmList.size() == 1){
			return (BinaryTreeNode)confirmList.get(0);
		}

		// If none of the nodes had a descendant match, or more than one confirmed, return closest node
		if (confirmList.size() == 0){
			confirmList = matchList;
		}
		
		double distance, closestDistance = -1;
		BinaryTreeNode closestMatch = null;
		for (int j = 0; j < confirmList.size(); j++){
			testNode = confirmList.get(j);
			// Check whether nodes' parents match

			distance = getDistance(testNode, goldNode);
			if (closestDistance == -1 || distance < closestDistance){
				closestDistance = distance;
				closestMatch = testNode;
			}
		}
		return closestMatch;

	}
	
	private BinaryTreeNode getNMPTerminationMatch(
			BinaryTreeNode goldNode, List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		
		// Assemble termination node's ancestor path
		Set<BinaryTreeNode> ancestors = new LinkedHashSet<BinaryTreeNode>();
		BinaryTreeNode node = goldNode.getParent();
		do{
			ancestors.add(node);
			node = node.getParent();
		} while (node != null);
		
		// Find nearest node to parent
		List<BinaryTreeNode> matchList = findNearestExtendedTerminationMatches(goldNode, testNodeList);
		BinaryTreeNode testNode, bestMatch = null, initialNode;
		
		if (writeDetails) System.out.println(matchList.size()+" matches");
		// Check whether nearest node(s) has parent that matches one of gold node's parents
		
		// Remove any nodes in matchList that have an ancestor in the matchList
		for (int i = 0; i < matchList.size(); i++){
			initialNode = (BinaryTreeNode)matchList.get(i);
			testNode = initialNode.getParent();
			while (testNode != null){
				if (matchList.contains(testNode)){
					matchList.remove(initialNode);
					i--;
					break;
				}
				testNode = testNode.getParent();
			}
		}
		
		double percentDifference, lowestDifference = -1;
		// Find the closest remaining match
		for (int i = 0; i < matchList.size(); i++){
			testNode = (BinaryTreeNode)matchList.get(i);

			// Attempt to match via normal means
			percentDifference = getMatchPathLengthDifference(goldNode, testNode, goldNodeList, testNodeList);

			// If no match, try matching accounting for rosettes
			if (percentDifference == 1){
				percentDifference = getNMPTerminationMatchPathDifference(goldNode, testNode, goldNodeList, testNodeList);
			}
			
			// Check whether nodes' paths match
			if (percentDifference < 1){
				if (writeDetails) System.out.println(getPositionString(testNode)+" matches");
				// If there are multiple matches, take the one with the smallest difference between path lengths
				if (lowestDifference == -1 || percentDifference < lowestDifference){
					lowestDifference = percentDifference;
					bestMatch = testNode;
				}
			}
		}
		if (bestMatch != null && writeDetails) 
			System.out.println("Closest is "+getPositionString(bestMatch)
				+" with percent path length difference of "+lowestDifference);

		// Remove all nodes below the gold termination from test list
		if (bestMatch != null && bestMatch.hasChildren()){
			LinkedStack stack = new LinkedStack();
			stack.push(bestMatch.getRight());
			stack.push(bestMatch.getLeft());
			while (!stack.isEmpty()){
				testNode = (BinaryTreeNode)stack.pop();
				matches.put(testNode, goldNode);
				if (testNode.hasChildren()){
					stack.push(testNode.getLeft());
					stack.push(testNode.getRight());
				}
			}
		}
		
		return bestMatch;
	}
	
	/**
	 * 
	 * @param goldNode
	 * @param testNode
	 * @param goldNodeList
	 * @param testNodeList
	 * @return Path distance difference percentage if swc continuation is found within threshold of the gold termination
	 *         and the path lengths are within threshold, otherwise 1.
	 */
	private double getNMPTerminationMatchPathDifference(
			BinaryTreeNode goldNode, BinaryTreeNode testNode, 
			List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		if (testNode.getParent() == null){
			return 1;
		}
		
		// Find nearest continuation points in last span within threshold before the first parent bifurcation 
		// Get treeNode based on test binary tree node
		SwcTreeNode treeNode = getTestTreePathNode(testNode);
		double runningPathLength = 0, selectedPathLength = 0, nodeDistance, smallestLocalDistance = -1;
		boolean found = false, wasOutsideThreshold = true;
		
		// Keep going until bifurcation is found
		do {
			// Determine distance between swc node and gold termination
			nodeDistance = SwcDataUtils.getDistance(treeNode.getSwcData(),goldNode.getSwcData());

			// If nodes are within threshold
			if (isWithinThreshold(treeNode.getSwcData(), goldNode.getSwcData())){
				found = true;
				// If this node is the first in a span, reset smallest local distance
				if (wasOutsideThreshold || nodeDistance <= smallestLocalDistance){
					smallestLocalDistance = nodeDistance;
					selectedPathLength = runningPathLength;
				}
				wasOutsideThreshold = false;
			}
			// Outside of threshold range, smallestLocalDistance = -1 as a flag
			else{
				wasOutsideThreshold = true;
			}

			runningPathLength += SwcDataUtils.getDistance(treeNode.getSwcData(), treeNode.getSwcParent().getSwcData());
			treeNode = treeNode.getSwcParent();
		} while (treeNode.getChildren().size() < 2);
		
		// If no continuation was within threshold, return the failure value
		if (!found){
			return 1;
		}
		
		// Get distance between nodes to use in path length check
		double difference, percentDifference;

		// Get path length at each node
		SwcDataNode goldData = (SwcDataNode)goldNode.getData();
		double goldPathLength = goldData.getSecondaryData().getPathLength();
		SwcDataNode testData = (SwcDataNode)testNode.getData();
		// Get test path length, accounting for distance from node to matched continuation point
		double testPathLength = testData.getSecondaryData().getPathLength() - selectedPathLength;
		
		// Start with parents of original nodes
		goldNode = goldNode.getParent();
		testNode = testNode.getParent();

		Set<BinaryTreeNode> checkedNodes = new HashSet<BinaryTreeNode>();
		List<BinaryTreeNode> nearbyNodes;
		boolean notDone = true;
		while (notDone){
			if (writeDetails){
				System.out.println("Looking at "+getPositionString(goldNode)+" and "+getPositionString(testNode));
				System.out.println("Baseline path length: "+goldPathLength);
				System.out.println("Sample path length: "+testPathLength);
			}
			// If ancestors are within euclidean distance threshold, see if path length matches
			if (isWithinThreshold(goldNode, testNode)){
				// Find path length difference
				difference = Math.abs(goldPathLength - testPathLength);
				// Subtract out error (distance between ancestor nodes)
				difference -= getDistance(goldNode, testNode);
				// Not factoring distance between continuation and gold termination
				
				percentDifference = difference/goldPathLength;
				if (writeDetails) System.out.println("Difference: "+percentDifference);
				if (percentDifference < XYPathErrorThreshold){
					return percentDifference;
				}
				// Node position matches, but path length doesn't. This is a miss
				else{
					return 1;
				}
				// Otherwise keep trying, as the "match" by location may have been wrong
			}

			else {
				// Decide which node to traverse up next
				if (goldPathLength < testPathLength){
					// If at root OR if ancestor has a known match and it hasn't been tried, then this node is a miss
					/*if (goldNode.getParent() == null ||
							(matches.containsKey(goldNode) && testNode != matches.get(goldNode))){*/
					// If at root, done (not stopping in other cases. false positives better than false negatives)
					if (goldNode.getParent() == null){
						notDone = false;
					}
					else{
						// Before moving on, check whether node matches a previously tested node
						nearbyNodes = findNearestNodes(goldNode, testNodeList, false);
						for (int i = 0; i < nearbyNodes.size(); i++){
							if (checkedNodes.contains(nearbyNodes.get(i))){
								// Node match found, path length by definition has not checked out
								notDone = false;
								break;
							}
						}

						// Add node to checked nodes
						checkedNodes.add(goldNode);
						// Update path length
						goldPathLength += ((SwcDataNode)goldNode.getData()).getSecondaryData().getPathLength();
						// Climb to next parent
						goldNode = goldNode.getParent();
					}
				}
				// Traversing up test/test path
				else{
					// If at root OR if ancestor has a known match and it hasn't been tried, then this node is a miss
					if (testNode.getParent() == null){
						notDone = false;
					}
					else{
						// Before moving on, check whether node matches a previously tested node
						nearbyNodes = findNearestNodes(testNode, goldNodeList, false);
						for (int i = 0; i < nearbyNodes.size(); i++){
							if (checkedNodes.contains(nearbyNodes.get(i))){
								// Node match found, path length by definition has not checked out
								notDone = false;
								break;
							}
						}

						// Add node to checked nodes
						checkedNodes.add(testNode);
						// Update path length
						testPathLength += ((SwcDataNode)testNode.getData()).getSecondaryData().getPathLength();
						// Climb to next parent
						testNode = testNode.getParent();
					}
				}
			}
		}

		// No match
		return 1;
	}
	
	/**
	 * Gets termination matches assuming those matches can be found at a greater distance from the target termination
	 * but that the path of the match must go within threshold distance of the target. Depending on the terminalPathThreshold
	 * the test branch might be shorter than the target node's branch. If so, the target's branch path must at some point
	 * fall within the threshold distance of the potential test match. Path length error is then tested based on the paths
	 * starting at the node of the shorter path and the closest point of the longer path. 
	 */
	private BinaryTreeNode getExtendedTerminationMatch(
			BinaryTreeNode goldNode, List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		
		// Find nearest test nodes to target
		List<BinaryTreeNode> matchList = findNearestExtendedTerminationMatches(goldNode, testNodeList);
		BinaryTreeNode testNode, bestMatch = null;
		
		if (writeDetails) System.out.println(matchList.size()+" initial matches");

		double percentError, lowestError = -1;

		// Keep track of the nodes whose paths do not pass by the target terminal node
		List<BinaryTreeNode> potentialShorterMatches = new ArrayList<BinaryTreeNode>();
		SwcTreeNode testPathNode, goldPathNode;
		TerminationMatchResult result;
		
		// Confirm the matches (based on path) and pick the best
		for (int i = 0; i < matchList.size(); i++){
			testNode = matchList.get(i);

			// Determine whether the potential match's path passes by the target node
			testPathNode = getPathPassingThroughNode(goldNode, testNode, true);
			if (testPathNode == null){
				/* This node does not lead to a path that passes through the target, so consider it as a node by which
				the gold standard path might pass through */
				potentialShorterMatches.add(testNode);
			}
			else{
				// Check whether nodes' paths match
				result = getExtendedTerminationMatchPathError(
						goldNode, testNode, testPathNode, goldNodeList, testNodeList);
				percentError = result.getError();

				if (percentError < 1){
					if (writeDetails) System.out.println(getPositionString(testNode)+" matches");
					// If there are multiple matches, take the one with the smallest difference between path lengths
					if (lowestError == -1 || percentError < lowestError){
						lowestError = percentError;
						bestMatch = testNode;
					}
				}
			}
		}
		/* If no match has been found AND if shorter test branches can be matched, 
		 * seek a match from those shorter branches
		 */
		if (bestMatch == null){
			for (int i = 0; i < potentialShorterMatches.size(); i++){
				testNode = potentialShorterMatches.get(i);

				// Determine whether the target's path passes by the potential match test node
				goldPathNode = getPathPassingThroughNode(testNode, goldNode, false);
				if (goldPathNode != null){
					// Check whether nodes' paths match
					result = getShortenedTerminationMatchPathError(
							goldNode, testNode, goldPathNode, goldNodeList, testNodeList);	
					percentError = result.getError();
					
					// If paths match AND test path is long enough based on path length threshold
					if (percentError < 1 
							&& result.getPathLengthChange() <= terminalPathThreshold){
						
						if (writeDetails) System.out.println(getPositionString(testNode)+" matches");
						// If there are multiple matches, take the one with the smallest difference between path lengths
						if (lowestError == -1 || percentError < lowestError){
							lowestError = percentError;
							bestMatch = testNode;
						}
					}
				}
			}
		}
		if (bestMatch != null && writeDetails) 
			System.out.println("Closest is "+getPositionString(bestMatch)
				+" with path length difference (as a proportion of threshold) of "+lowestError);

		// Remove all nodes below the gold termination from test list
		if (bestMatch != null && bestMatch.hasChildren()){
			LinkedStack stack = new LinkedStack();
			stack.push(bestMatch.getRight());
			stack.push(bestMatch.getLeft());
			while (!stack.isEmpty()){
				testNode = (BinaryTreeNode)stack.pop();
				matches.put(testNode, goldNode);
				if (testNode.hasChildren()){
					stack.push(testNode.getLeft());
					stack.push(testNode.getRight());
				}
			}
		}
		
		return bestMatch;
	}
	
	private List<BinaryTreeNode> findNearestExtendedTerminationMatches(
			BinaryTreeNode goldNode, List<BinaryTreeNode> testNodeList){
		List<BinaryTreeNode> nearestNodeList = new ArrayList<BinaryTreeNode>();
		BinaryTreeNode testNode, tmpNode;
		LinkedStack stack;
		boolean addNode;
	
		// Loop through test nodes
		for (int i = 0; i < testNodeList.size(); i++){
			testNode = testNodeList.get(i);
			// Gets distance between two nodes in space
			// If the distance is within the threshold, and this node hasn't been overused, add to list
			if (!matches.containsKey(testNode) && getDistance(goldNode, testNode) < terminalThreshold){
				if (writeDetails){
					SwcDataNode dat = testNode.getSwcData();
					System.out.println("Found node: "+dat.getX()+"; "+dat.getY()
							+"; XYDist: "+getXYDistance(goldNode, testNode)
							+"; ZDist: "+getZDistance(goldNode, testNode));
				}
				addNode = true;
				/* If this is the Neuromuscular Projection data set, exclude nodes that have children 
				 * outside the range of a rosette.
				 */
				if (dataset == DATASET_NM_PROJECTION && testNode.hasChildren()){
					stack = new LinkedStack();
					stack.push(testNode.getRight());
					stack.push(testNode.getLeft());
					while (!stack.isEmpty()){
						tmpNode = (BinaryTreeNode)stack.pop();
						if (getDistance(testNode, tmpNode) > terminalThreshold){
							addNode = false;
							break;
						}
					}
				}
				
				if (addNode){
					nearestNodeList.add(testNode);
				}
			}
		}
		return nearestNodeList;
	}

	// Find nearest continuation points in last span within threshold before the first parent bifurcation 
	private SwcTreeNode getPathPassingThroughNode(
			BinaryTreeNode targetNode, BinaryTreeNode traversingNode, boolean goldIsTarget){
		if (traversingNode.getParent() == null){
			return null;
		}
		
		if (debug || writeDetails) System.out.println("Getting path passing through node "+targetNode+"; traversingNode: "+traversingNode);
		
		// Get treeNode based on test binary tree node
		SwcTreeNode treeNode, closestTreeNode = null;
		if (goldIsTarget){
			treeNode = getTestTreePathNode(traversingNode);
		}
		else{
			treeNode = getGoldTreePathNode(traversingNode);
		}
		SwcSecondaryData pathLengthData = new SwcSecondaryData();
		double nodeDistance, smallestLocalDistance = -1;
		
		// Keep going until bifurcation is found
		do {
			//if (debug || writeDetails) System.out.println(treeNode.getSwcData());
			// Determine distance between swc node and gold termination
			nodeDistance = SwcDataUtils.getDistance(treeNode.getSwcData(),targetNode.getSwcData());

			// If nodes are within threshold
			if (isWithinThreshold(treeNode.getSwcData(), targetNode.getSwcData())){
				// If this node is closer to the target, reset smallest local distance
				if (nodeDistance <= smallestLocalDistance || smallestLocalDistance == -1){
					smallestLocalDistance = nodeDistance;
					closestTreeNode = treeNode;
					// Store the excess path length in the path node's path length member
					if (closestTreeNode.getSecondaryData() == null) closestTreeNode.setSecondaryData(new SwcSecondaryData()); 
					closestTreeNode.getSecondaryData().setPathLength(pathLengthData.getPathLength());
					closestTreeNode.getSecondaryData().setXYPathLength(pathLengthData.getXYPathLength());
					closestTreeNode.getSecondaryData().setZPathLength(pathLengthData.getZPathLength());
					if (debug || writeDetails) System.out.println("New closest node: "+nodeDistance);
				}
			}

			// Update the running path length (xy, z, and combined)
			pathLengthData.setPathLength(
					pathLengthData.getPathLength() +
					SwcDataUtils.getDistance(treeNode.getSwcData(), treeNode.getSwcParent().getSwcData()));
			pathLengthData.setXYPathLength(
					pathLengthData.getXYPathLength() +
					SwcDataUtils.getXYDistance(treeNode.getSwcData(), treeNode.getSwcParent().getSwcData()));
			pathLengthData.setZPathLength(
					pathLengthData.getZPathLength() +
					Math.abs(treeNode.getSwcData().getZ() - treeNode.getSwcParent().getSwcData().getZ()));
			treeNode = treeNode.getSwcParent();
		} while (treeNode.getChildren().size() < 2);
		
		return closestTreeNode;
	}

	private static class TerminationMatchResult{
		private double error;
		private double pathLengthChange;
		public TerminationMatchResult(double error, double pathLengthChange){
			this.error = error;
			this.pathLengthChange = pathLengthChange;
		}
		public TerminationMatchResult(double error){
			this.error = error;
		}
		public double getError() {
			return error;
		}
		public double getPathLengthChange() {
			return pathLengthChange;
		}
	}
	
	// The gold standard path is longer, therefore it is considered modified for the purposes of this function
	private TerminationMatchResult getShortenedTerminationMatchPathError(BinaryTreeNode goldNode, 
			BinaryTreeNode testNode, SwcTreeNode pathNode, List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		return getModifiedTerminationMatchPathError(goldNode, testNode, false, pathNode, goldNodeList, testNodeList);
	}
	
	// The test path is longer, therefore it is considered modified
	private TerminationMatchResult getExtendedTerminationMatchPathError(BinaryTreeNode goldNode, 
			BinaryTreeNode testNode, SwcTreeNode pathNode, List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		return getModifiedTerminationMatchPathError(goldNode, testNode, true, pathNode, goldNodeList, testNodeList);
	}

	
	/**
	 * @param goldNode
	 * @param testNode
	 * @param goldPathNode
	 * @param testPathNode
	 * @param goldNodeList
	 * @param testNodeList
	 * @return Path distance difference percentage if swc continuation is found within threshold of the gold termination
	 *         and the path lengths are within threshold, otherwise 1.
	 */
	
	private TerminationMatchResult getModifiedTerminationMatchPathError(
			BinaryTreeNode goldNode, BinaryTreeNode testNode, boolean testPathModified, 
			SwcTreeNode pathNode, List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		TerminationMatchResult result = new TerminationMatchResult(1);
		if (testNode.getParent() == null){
			return result;
		}
		BinaryTreeNode potentialMatch = testNode;
		SwcSecondaryData testSecondaryData = testNode.getSwcData().getSecondaryData();
		SwcSecondaryData goldSecondaryData = goldNode.getSwcData().getSecondaryData();
		
		// Get path length at each node
		SwcSecondaryData goldPathLength = new SwcSecondaryData();
		SwcSecondaryData testPathLength = new SwcSecondaryData();

		// Get test path length, accounting for distance from node to matched continuation point
		if (testPathModified){
			goldPathLength.setPathLength(goldSecondaryData.getPathLength());
			goldPathLength.setXYPathLength(goldSecondaryData.getXYPathLength());
			goldPathLength.setZPathLength(goldSecondaryData.getZPathLength());

			testPathLength.setPathLength(
					testSecondaryData.getPathLength() - pathNode.getSecondaryData().getPathLength());
			testPathLength.setXYPathLength( 
					testSecondaryData.getXYPathLength() - pathNode.getSecondaryData().getXYPathLength());
			testPathLength.setZPathLength(
					testSecondaryData.getZPathLength() - pathNode.getSecondaryData().getZPathLength());
		}
		else{
			goldPathLength.setPathLength(
					goldSecondaryData.getPathLength() - pathNode.getSecondaryData().getPathLength());
			goldPathLength.setXYPathLength( 
					goldSecondaryData.getXYPathLength() - pathNode.getSecondaryData().getXYPathLength());
			goldPathLength.setZPathLength(
					goldSecondaryData.getZPathLength() - pathNode.getSecondaryData().getZPathLength());

			testPathLength.setPathLength(testSecondaryData.getPathLength());
			testPathLength.setXYPathLength(testSecondaryData.getXYPathLength());
			testPathLength.setZPathLength(testSecondaryData.getZPathLength());
		}

		double testXYPathLength, testZPathLength, percentError;
		BinaryTreeNode goldTarget = goldNode;

		// Set up path length adjustments due to allowable position error
		boolean isBranchLeft = goldTarget.isLeft();
		EuclideanPoint descendantTrajectory;
		EuclideanPoint ancestorTrajectory;
		double xyDescendantAdjustment, zDescendantAdjustment;
		if (testPathModified){
			descendantTrajectory = goldNode.getSwcData().getSecondaryData().getParentTrajectoryPoint();
			xyDescendantAdjustment = getEndNodeXYDistanceDifference(
					goldTarget.getSwcData(), descendantTrajectory, pathNode.getSwcData());
			zDescendantAdjustment = getEndNodeZDistanceDifference(
					goldTarget.getSwcData(), descendantTrajectory, pathNode.getSwcData());
		}
		else{
			SwcDataNode thresholds = new SwcDataNode();
			thresholds.setX(XYThreshold);
			thresholds.setZ(ZThreshold);
			descendantTrajectory = ReadSWC.findParentTrajectory(pathNode, thresholds);
			xyDescendantAdjustment = getEndNodeXYDistanceDifference(
					pathNode.getSwcData(), descendantTrajectory, testNode.getSwcData());
			zDescendantAdjustment = getEndNodeZDistanceDifference(
					pathNode.getSwcData(), descendantTrajectory, testNode.getSwcData());
		}
		
		// Start with parents of original nodes
		goldNode = goldNode.getParent();
		testNode = testNode.getParent();

		/* Traverse upwards from target and potential match (continuation point) until a matching ancestor is found.
		 * Calculate path length error
		 */
		Set<BinaryTreeNode> checkedNodes = new HashSet<BinaryTreeNode>();
		boolean notDone = true;
		while (notDone){
			if (writeDetails){
				System.out.println("Looking at "+getPositionString(goldNode)+" and "+getPositionString(testNode));
				System.out.println("Baseline path length: "+goldPathLength.getPathLength());
				System.out.println("Sample path length: "+testPathLength.getPathLength());
			}
			// If ancestors are within Euclidean distance threshold, see if path length matches
			if (isWithinThreshold(goldNode, testNode)){
				// Get the ancestor trajectory for modifying test path length
				ancestorTrajectory = getAncestorTrajectory(goldNode, goldTarget, isBranchLeft);
				
				// Calculate test XY and Z path length modified by end node positions
				testXYPathLength = testPathLength.getXYPathLength() 
					+ getEndNodeXYDistanceDifference(goldNode.getSwcData(), ancestorTrajectory, testNode.getSwcData())
					+ xyDescendantAdjustment;
				testZPathLength = testPathLength.getZPathLength() 
					+ getEndNodeZDistanceDifference(goldNode.getSwcData(), ancestorTrajectory, testNode.getSwcData())
					+ zDescendantAdjustment;

				percentError = pathLengthMatches(goldPathLength, testXYPathLength, testZPathLength);
				if (percentError < 1 && writeDetails) if (writeDetails) System.out.println("Found match for termination: "+potentialMatch.getSwcData());
				// If the path length error is within threshold, "percentError" is the (weighted) average error
				// Otherwise it is 1, signifying that the error is too large and a miss
				result = new TerminationMatchResult(percentError, pathNode.getSecondaryData().getPathLength());
				return result;
			}
			else {
				// Decide which node to traverse up next
				if (goldPathLength.getPathLength() < testPathLength.getPathLength()){
					// If at root OR if ancestor has a known match and it hasn't been tried, then this node is a miss
					goldNode = climbTree(goldNode, testNodeList, checkedNodes, goldPathLength);
					if (goldNode == null){
						return result;
					}
				}
				// Traversing up test path
				else{
					// If at root OR if ancestor has a known match and it hasn't been tried, then this node is a miss
					testNode = climbTree(testNode, goldNodeList, checkedNodes, testPathLength);
					if (testNode == null){
						return result;
					}
				}
			}
		}

		// No match
		return result;
	}

	private BinaryTreeNode climbTree(BinaryTreeNode node, 
			List<BinaryTreeNode> otherTreeNodeList, Set<BinaryTreeNode> checkedNodes, SwcSecondaryData pathLength){
		// If at root, done (not stopping in other cases. false positives better than false negatives)
		if (node.getParent() == null){
			return null;
		}
		else{
			// Before moving on, check whether gold ancestor matches a previously tested node
			List<BinaryTreeNode> nearbyNodes = findNearestNodes(node, otherTreeNodeList, false);
			for (int i = 0; i < nearbyNodes.size(); i++){
				if (checkedNodes.contains(nearbyNodes.get(i))){
					// Node match found, path length by definition has not checked out
					return null;
				}
			}

			// Add ancestor node to checked nodes
			checkedNodes.add(node);
			// Update path length
			pathLength.setPathLength(
					pathLength.getPathLength() +
					((SwcDataNode)node.getData()).getSecondaryData().getPathLength());
			pathLength.setXYPathLength(
					pathLength.getXYPathLength() +
					((SwcDataNode)node.getData()).getSecondaryData().getXYPathLength());
			pathLength.setZPathLength(
					pathLength.getZPathLength() +
					((SwcDataNode)node.getData()).getSecondaryData().getZPathLength());
			// Climb to next parent
			return node.getParent();
		}
	}
	
	private SwcTreeNode getTestTreePathNode(BinaryTreeNode node){
		return (SwcTreeNode)testTreePathNodes.get(new Integer(node.getSwcData().getNodeId()));
	}
	private SwcTreeNode getGoldTreePathNode(BinaryTreeNode node){
		return (SwcTreeNode)goldTreePathNodes.get(new Integer(node.getSwcData().getNodeId()));
	}

	private EuclideanPoint getAncestorTrajectory(BinaryTreeNode ancestor, BinaryTreeNode descendant,
			boolean isBranchLeft){
		EuclideanPoint ancestorTrajectory;
		if (isBranchLeft){
			ancestorTrajectory = ancestor.getSwcData().getSecondaryData().getLeftTrajectoryPoint();
		}
		else{
			ancestorTrajectory = ancestor.getSwcData().getSecondaryData().getRightTrajectoryPoint();
		}
		// If the ancestor has a child within threshold, a path specific trajectory must be calculated
		if (ancestorTrajectory.getX() == ReadSWC.TRAJECTORY_NONE || ancestorTrajectory.getZ() == ReadSWC.TRAJECTORY_NONE){
			EuclideanPoint tmpTrajectory = getTrajectoryForPath(ancestor, descendant);
			if (ancestorTrajectory.getX() == ReadSWC.TRAJECTORY_NONE){
				ancestorTrajectory.setX(tmpTrajectory.getX());
				ancestorTrajectory.setY(tmpTrajectory.getY());
			}
			if (ancestorTrajectory.getZ() == ReadSWC.TRAJECTORY_NONE){
				ancestorTrajectory.setZ(tmpTrajectory.getZ());
			}
			if (writeDetails) System.out.println("Getting trajectory for path with ancestor: "+getPositionString(ancestor)+"; descendant: "+getPositionString(descendant)+": "+getPositionString(ancestorTrajectory));
		}
		return ancestorTrajectory;
	}
	
	/**
	 * 
	 * @param goldNode
	 * @param testNodeList
	 * @return
	 * Whether or not a node is a continuation in the test tree
	 * Find first ancestor known to be in test tree 
	 * (also consider ancestor's of known ancestor that are within threshold distance of the first known ancestor).
	 * Search down from gold node for matches (within dist threshold), then compare path lengths
	 */
	private boolean isContinuation(BinaryTreeNode goldNode, List<BinaryTreeNode> testNodeList){
		return isContinuation(goldNode, testNodeList, true);
	}
	private boolean isContinuation(BinaryTreeNode goldNode, List<BinaryTreeNode> testNodeList, boolean addToLists){
		// Terminal nodes can't be continuations
		if (!goldNode.hasChildren()){
			return false;
		}
		List<BinaryTreeNode> ancestorNodeMatches;
		
		// Get path length at each node
		SwcSecondaryData goldPathLength = new SwcSecondaryData();

		// Find first parent of gold node that has a match
		BinaryTreeNode ancestorMatch = null, ancestorNode = goldNode.getParent();
		goldPathLength.setPathLength(goldNode.getSwcData().getSecondaryData().getPathLength());
		goldPathLength.setXYPathLength(goldNode.getSwcData().getSecondaryData().getXYPathLength());
		goldPathLength.setZPathLength(goldNode.getSwcData().getSecondaryData().getZPathLength());
		boolean isMatch = false, matchedAncestor = false, isBranchLeft = goldNode.isLeft();
		
		// Look for match until one is found or root is reached
		while (ancestorNode != null && !matchedAncestor) {
			// If known match, use it (and don't keep looking)
			if (writeDetails) System.out.println("Checking ancestor: "+ancestorNode.getSwcData());
			if (matches.containsKey(ancestorNode)){
				if (writeDetails) System.out.println("Known ancestor node");
				ancestorMatch = matches.get(ancestorNode);
				if (isContinuation(goldNode, ancestorNode, ancestorMatch, isBranchLeft, 
						goldPathLength, testNodeList, addToLists)){
					return true;
				}
				/* If known ancestor does not provide a path for a continuation, try successive ancestors that are within
				threshold distance of the current ancestor (next loop), but stop searching after that */
				matchedAncestor = true;
			}
			// While no match was made, that might just be because the node's connection to its parent was wrong
			ancestorNodeMatches = findNearestNodes(ancestorNode, testNodeList);
			for (int i = 0; i < ancestorNodeMatches.size() && !isMatch; i++){
				ancestorMatch = ancestorNodeMatches.get(i);
				isMatch = isContinuation(goldNode, ancestorNode, ancestorMatch, isBranchLeft, 
						goldPathLength, testNodeList, addToLists);
				// If we've found the match, let's get out of here
				if (isMatch) return true;
			}

			// No match, keep going
			goldPathLength.incrementPathLength(ancestorNode.getSwcData().getSecondaryData().getPathLength());
			goldPathLength.incrementXYPathLength(ancestorNode.getSwcData().getSecondaryData().getXYPathLength());
			goldPathLength.incrementZPathLength(ancestorNode.getSwcData().getSecondaryData().getZPathLength());
			isBranchLeft = ancestorNode.isLeft();
			ancestorNode = ancestorNode.getParent();
		}

		// Found no match, so this isn't a continuation
		return false;
	}

	// Subroutine for overall isContinuation function
	private boolean isContinuation(BinaryTreeNode goldNode, BinaryTreeNode ancestorNode,
			BinaryTreeNode ancestorMatch, boolean isLeftBranch, SwcSecondaryData goldPathLength, 
			List<BinaryTreeNode> testNodeList, boolean addToLists){

		// If there is no parent match, this cannot be confirmed as a continuation
		if (ancestorMatch == null) return false;
		
		if (writeDetails) System.out.println("Ancestor Match, Gold: "+getPositionString(ancestorNode)+"; test: "+getPositionString(ancestorMatch));
		
		Map<BinaryTreeNode,SwcSecondaryData> pathLengthMap = new HashMap<BinaryTreeNode,SwcSecondaryData>();
		pathLengthMap.put(goldNode, goldPathLength);
		
		BinaryTreeNode leftChildMatch, rightChildMatch;
		EuclideanPoint ancestorTrajectory;
		if (isLeftBranch){
			ancestorTrajectory = ancestorNode.getSwcData().getSecondaryData().getLeftTrajectoryPoint();
		}
		else{
			ancestorTrajectory = ancestorNode.getSwcData().getSecondaryData().getRightTrajectoryPoint();
		}
		
		// Search down either side from gold node for match(es)
		leftChildMatch = getDescendantInForContinuation(
				goldNode.getLeft(), ancestorNode, ancestorMatch, ancestorTrajectory, pathLengthMap, testNodeList);
		rightChildMatch = getDescendantInForContinuation(
				goldNode.getRight(), ancestorNode, ancestorMatch, ancestorTrajectory, pathLengthMap, testNodeList);

		// If both are found, then there is a common node outside of the threshold range
		if (leftChildMatch != null && rightChildMatch != null){
			BinaryTreeNode commonAncestor = leastCommonAncestor(leftChildMatch, rightChildMatch, ancestorMatch);
			// TODO: Decide whether this is an appropriate determinant for a "distant match" or if it is even meaningful
			if (commonAncestor != null && isWithinDistantMatchThreshold(commonAncestor, goldNode)){
				matches.put(goldNode, commonAncestor);
				matches.put(commonAncestor, goldNode);
				if (writeDetails)
					System.out.println("Distant match to "+getPositionString(commonAncestor));
				if (addToLists){
					distantMatches.add(goldNode);
				}
			}
			else if (addToLists){
				continuations.add(goldNode);
			}
			return true;
		}
		// Return true if either child's side gets a match
		else if (leftChildMatch != null || rightChildMatch != null){
			if (addToLists){
				continuations.add(goldNode);
			}
			return true;
		}
		return false;
	}
	
	private BinaryTreeNode getDescendantInForContinuation(
			BinaryTreeNode firstNode, BinaryTreeNode ancestorNode,
			BinaryTreeNode ancestorMatch, EuclideanPoint ancestorTrajectory,
			Map<BinaryTreeNode,SwcSecondaryData> pathLengthMap, List<BinaryTreeNode> testNodeList){
		BinaryTreeNode childMatch = null, tmpNode, goldNode;
		double testXYPathLength, testZPathLength;
		boolean done;
		SwcSecondaryData goldPathLength, prevPathLength, testPathLength;
		EuclideanPoint specificAncestorTrajectory = ancestorTrajectory, descendantTrajectory;
		List<BinaryTreeNode> testMatches;
		
		LinkedStack stack = new LinkedStack();
		stack.push(firstNode);
		// Search down from gold node for matches
		while (!stack.isEmpty()){
			goldNode = (BinaryTreeNode)stack.pop();
			if (writeDetails) System.out.println("Descendant "+getPositionString(goldNode.getSwcData()));
			descendantTrajectory = goldNode.getSwcData().getSecondaryData().getParentTrajectoryPoint();

			// If there is a known match for the descendant, use it
			if (matches.containsKey(goldNode)){
				testMatches = new ArrayList<BinaryTreeNode>();
				testMatches.add(matches.get(goldNode));
			}
			else{
				// Find nodes within threshold of descendant
				testMatches = findNearestNodes(goldNode, testNodeList, false);
			}
			
			// Calculate and score gold path length
			prevPathLength = pathLengthMap.get(goldNode.getParent());
			goldPathLength = new SwcSecondaryData();
			goldPathLength.setPathLength(
				goldNode.getSwcData().getSecondaryData().getPathLength()
				+ prevPathLength.getPathLength());
			goldPathLength.setXYPathLength(
				goldNode.getSwcData().getSecondaryData().getXYPathLength()
				+ prevPathLength.getXYPathLength());
			goldPathLength.setZPathLength(
				goldNode.getSwcData().getSecondaryData().getZPathLength()
				+ prevPathLength.getZPathLength());
			pathLengthMap.put(goldNode, goldPathLength);
			
			for (int i = 0; i < testMatches.size(); i++){
				childMatch = testMatches.get(i);
				if (writeDetails) System.out.println("Matching on gold child: "+getPositionString(goldNode)+"; test match: "+ getPositionString(childMatch));
				
				// Go up to parent (or node with path length calculated) to find path length
				testPathLength = new SwcSecondaryData();
				testPathLength.addData(childMatch.getSwcData().getSecondaryData());
				tmpNode = childMatch.getParent();

				// Don't bother tracing up if child match's parent is ancestor
				done = (tmpNode == ancestorMatch);
				while (!done){
					if (pathLengthMap.containsKey(tmpNode)){
						prevPathLength = pathLengthMap.get(tmpNode);
						testPathLength.incrementPathLengths(prevPathLength);
						done = true;
					}
					else{
						prevPathLength = (SwcSecondaryData)pathLengthMap.get(tmpNode);
						testPathLength.incrementPathLengths(tmpNode.getSwcData().getSecondaryData());
						tmpNode = tmpNode.getParent();
						if (tmpNode == null){
							// Path not found: matched to descendant was ancestor of matched ancestor
							done = true;
						}
					}
					if (tmpNode == ancestorMatch){
						done = true;
					}
				}
				
				// Only continue if the descendant match is the descendant of the ancestor match
				if (tmpNode == null){
					if (writeDetails) System.out.println("Descendant match not a descendant of ancestor match");
				}
				else{
					pathLengthMap.put(childMatch, testPathLength);
					if (ancestorTrajectory.getX() == ReadSWC.TRAJECTORY_NONE || ancestorTrajectory.getZ() == ReadSWC.TRAJECTORY_NONE){
						specificAncestorTrajectory = getTrajectoryForPath(ancestorNode, goldNode);
						if (ancestorTrajectory.getX() != ReadSWC.TRAJECTORY_NONE){
							specificAncestorTrajectory.setX(ancestorTrajectory.getX());
							specificAncestorTrajectory.setY(ancestorTrajectory.getY());
						}
						if (ancestorTrajectory.getZ() != ReadSWC.TRAJECTORY_NONE){
							specificAncestorTrajectory.setZ(ancestorTrajectory.getZ());
						}
					}

					if (writeDetails) System.out.println("Path distances: "+goldPathLength.getPathLength()+" "+testPathLength.getPathLength());
	
					// Calculate test XY and Z path length modified by end node positions
					testXYPathLength = testPathLength.getXYPathLength() 
						+ getEndNodeXYDistanceDifference(ancestorNode.getSwcData(), specificAncestorTrajectory, ancestorMatch.getSwcData())
						+ getEndNodeXYDistanceDifference(goldNode.getSwcData(), descendantTrajectory, childMatch.getSwcData());
					testZPathLength = testPathLength.getZPathLength() 
						+ getEndNodeZDistanceDifference(ancestorNode.getSwcData(), specificAncestorTrajectory, ancestorMatch.getSwcData())
						+ getEndNodeZDistanceDifference(goldNode.getSwcData(), descendantTrajectory, childMatch.getSwcData());

					if (writeDetails) System.out.println("XY Test Path: "+testPathLength.getXYPathLength()+"; ancestorAdjust: "+
							getEndNodeXYDistanceDifference(ancestorNode.getSwcData(), specificAncestorTrajectory, ancestorMatch.getSwcData())+
							"; descendantAdjust: "+ getEndNodeXYDistanceDifference(goldNode.getSwcData(), descendantTrajectory, childMatch.getSwcData()));
					
					if (pathLengthMatches(goldPathLength, testXYPathLength, testZPathLength) < 1){
						if (writeDetails) System.out.println("Found child match for continuation: "+childMatch.getData());
						return childMatch;
					}
					// Otherwise keep trying, as the "match" by location may have been wrong
					else if (writeDetails) System.out.println("Try another.");
				}
			}
			// Add child nodes to stack only if this node didn't already have a known match
			if (goldNode.hasChildren() && !matches.containsKey(goldNode)){
				stack.push(goldNode.getLeft());
				stack.push(goldNode.getRight());
			}
		}
		return null;
	}
	
	/**
	 * 
	 * @param goldPathLength
	 * @param xyTestPathLength
	 * @param zTestPathLength
	 * @return Average XY and Z path length error as a proportion of the threshold
	 */
	private double pathLengthMatches(SwcSecondaryData goldPathLength, 
			double xyTestPathLength, double zTestPathLength){
		
		double xyLocalPathErrorThreshold = XYPathErrorThreshold,
			zLocalPathErrorThreshold = ZPathErrorThreshold;
		
		double xyDiff = Math.abs(goldPathLength.getXYPathLength() - xyTestPathLength);
		double xyErr = xyDiff/goldPathLength.getPathLength();
		if (writeDetails) System.out.println("goldXYPath: "+goldPathLength.getXYPathLength()+"; testXYPath: " +xyTestPathLength);
		if (writeDetails) System.out.println("xyDiff = "+xyDiff+"; xyErr: "+xyErr);
		
		double zDiff = Math.abs(goldPathLength.getZPathLength() - zTestPathLength);
		double zErr = zDiff/goldPathLength.getPathLength();
		if (writeDetails) System.out.println("goldZPath: "+goldPathLength.getZPathLength()+"; testZPath: " +zTestPathLength);
		if (writeDetails) System.out.println("zDiff = "+zDiff+"; zErr = "+zErr);

		if (goldPathLength.getXYPathLength() < XYThreshold){
			// If both paths are within the euclidean threshold distance, no check needed 
			if (xyTestPathLength < XYThreshold){
				xyErr = 0;
			}
			// If gold standard path is within, but test path is not, increase the path length error threshold
			else{
				xyLocalPathErrorThreshold = LocalPathErrorThreshold;
			}
		}
		if (goldPathLength.getZPathLength() < ZThreshold){
			// If both paths are within the euclidean threshold distance, no check needed 
			if (zTestPathLength < ZThreshold){
				zErr = 0;
			}
			// If gold standard path is within, but test path is not, increase the path length error threshold
			else{
				zLocalPathErrorThreshold = LocalPathErrorThreshold;
			}
		}
		
		if (xyErr < xyLocalPathErrorThreshold && zErr < zLocalPathErrorThreshold){
			return (xyErr/xyLocalPathErrorThreshold + zErr/zLocalPathErrorThreshold)/2;
		}
		else{
			return 1;
		}
	}

	private BinaryTreeNode leastCommonAncestor(
			BinaryTreeNode node1, BinaryTreeNode node2, BinaryTreeNode knownCommonAncestor){
		List<BinaryTreeNode> node1List = new ArrayList<BinaryTreeNode>();
		
		if (writeDetails){
			System.out.println("Known common ancestor: "+getPositionString(knownCommonAncestor));
		}
		// Assemble node1's ancestor list
		node1 = node1.getParent();
		while (node1 != knownCommonAncestor){
			if (writeDetails) System.out.println("Node1: "+getPositionString(node1));
			node1List.add(node1);
			node1 = node1.getParent();
		}

		// Find first ancestor node of node2 that is an ancestor node of node1
		node1 = node2.getParent();
		while (node2 != knownCommonAncestor && node2 != null){
			if (node1List.contains(node2)){
				return node2;
			}
			node2 = node2.getParent();
		}
		
		return null;
	}
	
	/**
	 * 
	 * @param goldNode
	 * @param testNodeList
	 * @return list of nearest nodes
	 */
	private List<BinaryTreeNode> findNearestNodes(BinaryTreeNode goldNode, List<BinaryTreeNode> testNodeList){
		return findNearestNodes(goldNode, testNodeList, true);
	}
	private List<BinaryTreeNode> findNearestNodes(BinaryTreeNode goldNode, List<BinaryTreeNode> testNodeList, boolean checkPreviousUse){
		List<BinaryTreeNode> nearestNodeList = new ArrayList<BinaryTreeNode>();
		BinaryTreeNode testNode;

		// Loop through test nodes
		for (int i = 0; i < testNodeList.size(); i++){
			testNode = testNodeList.get(i);
			// Gets distance between two nodes in space
			// If the distance is within the threshold, and this node hasn't been overused, add to list
			if (((!checkPreviousUse || !matches.containsKey(testNode))) && isWithinThreshold(goldNode, testNode)){
				if (writeDetails){
					SwcDataNode dat = (SwcDataNode)testNode.getData();
					System.out.println("Found node: "+dat.getX()+"; "+dat.getY()
							+"; XYDist: "+getXYDistance(goldNode, testNode)
							+"; ZDist: "+getZDistance(goldNode, testNode));
				}
				nearestNodeList.add(testNode);
			}
		}

		return nearestNodeList;
	}

	// Determine whether the node is a match or not
	private boolean nodeMatches(BinaryTreeNode goldNode, BinaryTreeNode testNode, 
			List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		double lengthDiff = getMatchPathLengthDifference(goldNode, testNode, goldNodeList, testNodeList);
		if (lengthDiff < 1) 
			return true; 
		else 
			return false;
	}
	/** 
	 * 
	 * @param goldNode
	 * @param testNode
	 * @param goldNodeList
	 * @param testNodeList
	 * @return Either the path length percent difference (XY and Z included) if XY and Z length error rates are within threshold,
	 * or 1 if they are not within threshold
	 * 
	 */ 
	private double getMatchPathLengthDifference(BinaryTreeNode goldNode, BinaryTreeNode testNode, 
			List<BinaryTreeNode> goldNodeList, List<BinaryTreeNode> testNodeList){
		if (testNode.getParent() == null){
			return 1;
		}

		// Used for determining path length threshold based on relative amounts of XY vs Z error
		BinaryTreeNode goldTarget = goldNode;

		// Get path length at each node
		SwcSecondaryData goldPathLength = new SwcSecondaryData();
		goldPathLength.incrementPathLengths(goldNode.getSwcData().getSecondaryData());
		SwcSecondaryData testPathLength = new SwcSecondaryData();
		testPathLength.incrementPathLengths(testNode.getSwcData().getSecondaryData());
		
		SwcDataNode goldData;
		EuclideanPoint ancestorTrajectory;
		double testXYPathLength, testZPathLength, percentError;

		// Calculate difference in path length for test path based on distance from gold node
		double testPathXYMod = getEndNodeXYDistanceDifference(
				goldNode.getSwcData(), goldNode.getSwcData().getSecondaryData().getParentTrajectoryPoint(), testNode.getSwcData());
		// Since this path modification will persist for all ancestor checks, it can be stored directly in the continually updated path length
		testPathLength.incrementXYPathLength(testPathXYMod);
		double testPathZMod = getEndNodeZDistanceDifference(
				goldNode.getSwcData(), goldNode.getSwcData().getSecondaryData().getParentTrajectoryPoint(), testNode.getSwcData());
		testPathLength.incrementZPathLength(testPathZMod);

		if (writeDetails) System.out.println("Test Path Mods (XY,Z): ("+testPathXYMod+", "+testPathZMod+"); TrajecoryPt: "+
				getPositionString(goldNode.getSwcData().getSecondaryData().getParentTrajectoryPoint()));
		
		// Start with parents of original nodes, climb using goldNode and testNode
		boolean isBranchLeft = goldNode.isLeft();
		goldNode = goldNode.getParent();
		testNode = testNode.getParent();

		Set<BinaryTreeNode> checkedNodes = new HashSet<BinaryTreeNode>();
		List<BinaryTreeNode> nearbyNodes;
		boolean noMatch = true, notDone = true;
		while (noMatch && notDone){
			if (writeDetails || debug){
				System.out.println("Looking at ancestor "+getPositionString(goldNode)+" and "+getPositionString(testNode));
			}
			// If the gold standard and test ancestors are within threshold distance, confirm match by path length
			if (isWithinThreshold(goldNode, testNode)){
				goldData = goldNode.getSwcData();
				// Get the ancestor trajectory for modifying test path length
				if (isBranchLeft){
					ancestorTrajectory = goldData.getSecondaryData().getLeftTrajectoryPoint();
				}
				else{
					ancestorTrajectory = goldData.getSecondaryData().getRightTrajectoryPoint();
				}
				// If the ancestor has a child within threshold, a path specific trajectory must be calculated
				if (ancestorTrajectory.getX() == ReadSWC.TRAJECTORY_NONE || ancestorTrajectory.getZ() == ReadSWC.TRAJECTORY_NONE){
					EuclideanPoint tmpTrajectory = getTrajectoryForPath(goldNode, goldTarget);
					if (ancestorTrajectory.getX() == ReadSWC.TRAJECTORY_NONE){
						ancestorTrajectory.setX(tmpTrajectory.getX());
						ancestorTrajectory.setY(tmpTrajectory.getY());
					}
					if (ancestorTrajectory.getZ() == ReadSWC.TRAJECTORY_NONE){
						ancestorTrajectory.setZ(tmpTrajectory.getZ());
					}
					if (writeDetails) System.out.println("Getting trajectory for path with ancestor: "+getPositionString(goldNode)+"; descendant: "+getPositionString(goldTarget)+": "+getPositionString(ancestorTrajectory));
				}
				
				if (writeDetails) System.out.println("Gold path length: "+goldPathLength.getPathLength()+"; Test path length: "+testPathLength.getPathLength()
						+"; TrajectoryPt: "+ ancestorTrajectory);

				// Calculate test XY and Z path length modified by end node positions
				// temporary variables used because this modification will change with each successive ancestor 
				testXYPathLength = testPathLength.getXYPathLength() 
					+ getEndNodeXYDistanceDifference(goldData, ancestorTrajectory, testNode.getSwcData());
				testZPathLength = testPathLength.getZPathLength() 
					+ getEndNodeZDistanceDifference(goldData, ancestorTrajectory, testNode.getSwcData());

				percentError = pathLengthMatches(goldPathLength, testXYPathLength, testZPathLength);
				if (percentError < 1){
					return percentError;
				}
				else{
					return 1;
				}
				// Otherwise keep trying, as the "match" by location may have been wrong
			}
			else {
				// Decide which node to traverse up next
				if (goldPathLength.getPathLength() < testPathLength.getPathLength()){
					// If at root OR if ancestor has a known match and it hasn't been tried, then this node is a miss
					/*if (goldNode.getParent() == null ||
							(matches.containsKey(goldNode) && testNode != matches.get(goldNode))){*/
					// If at root, done (not stopping in other cases. false positives better than false negatives)
					if (goldNode.getParent() == null){
						notDone = false;
					}
					else{
						// Before moving on, check whether node matches a previously tested node
						nearbyNodes = findNearestNodes(goldNode, testNodeList, false);
						for (int i = 0; i < nearbyNodes.size(); i++){
							if (checkedNodes.contains(nearbyNodes.get(i))){
								// Node match found, path length by definition has not checked out
								notDone = false;
								break;
							}
						}

						// Add node to checked nodes
						checkedNodes.add(goldNode);
						// Update path length
						goldPathLength.incrementPathLengths(goldNode.getSwcData().getSecondaryData());
						// Climb to next parent
						isBranchLeft = goldNode.isLeft();
						goldNode = goldNode.getParent();
					}
				}
				else{
					// If at root OR if ancestor has a known match and it hasn't been tried, then this node is a miss
					if (testNode.getParent() == null){
						notDone = false;
					}
					else{
						// Before moving on, check whether node matches a previously tested node
						nearbyNodes = findNearestNodes(testNode, goldNodeList, false);
						for (int i = 0; i < nearbyNodes.size(); i++){
							if (checkedNodes.contains(nearbyNodes.get(i))){
								// Node match found, path length by definition has not checked out
								notDone = false;
								break;
							}
						}

						// Add node to checked nodes
						checkedNodes.add(testNode);
						// Update path length
						testPathLength.incrementPathLengths(testNode.getSwcData().getSecondaryData());
						// Climb to next parent
						testNode = testNode.getParent();
					}
				}
			}
		}

		// No match, return 1
		return 1;
	}

	// Excess nodes weighed by degree of excess terms without any matches in between
	private double weighExcess(SwcTreeNode testRoot, List<BinaryTreeNode> goldTrees){
		int weightSum = 0;
		// Traverse gold tree and get a list of nodes
		List<BinaryTreeNode> goldNodeList = new ArrayList<BinaryTreeNode>();
		BinaryTreeNode rootNode;
		for (int i = 0; i < goldTrees.size(); i++){
			rootNode = goldTrees.get(i);
			goldNodeList.addAll(
					BinaryTreeUtils.createNodeList(rootNode));
		}

		BinaryTreeNode node;
		// Get all nodes into stack such that nodes will always pop before their parents
		LinkedStack setupStack, useStack = new LinkedStack();
		for (int i = 0; i < testRoot.getChildren().size(); i++){
			node = (BinaryTreeNode)testRoot.getChildren().get(i);
			setupStack = new LinkedStack();
			setupStack.push(node);
			while (!setupStack.isEmpty()){
				node = (BinaryTreeNode)setupStack.pop();
				if (node.hasChildren()){
					setupStack.push(node.getLeft());
					setupStack.push(node.getRight());
					// Purposefully not including root node
					useStack.push(node.getLeft());
					useStack.push(node.getRight());
				}
			}
		}
		
		List<BinaryTreeNode> nearbyNodes;
		int numExcessNodes;
		double excessWeight;
		Map<BinaryTreeNode,Integer> directTermExcess = new HashMap<BinaryTreeNode,Integer>();
		SwcDataNode data;

		// Determine whether test nodes are excess or not
		while (!useStack.isEmpty()){
			node = (BinaryTreeNode)useStack.pop();
			data = node.getSwcData();
			writeDetails = false;
			if (isWithinXYThreshold(xyCheck, data)){
				writeDetails = true;
				System.out.println("Excess Node Check X: "+data.getX()+"; Y: "+data.getY());
			}
			// Bifurcation node
			if (node.hasChildren()){
				if (writeDetails) System.out.println("Has children");
				// Accumulating terminal hits under this node
				numExcessNodes = directTermExcess.get(node.getLeft());
				numExcessNodes += directTermExcess.get(node.getRight());
				
				excessWeight = getExcessWeight(numExcessNodes, node, directTermExcess);
				
				/* Only add to the weight if this node is a miss, not a  
				   continuation, and if there are no unmatched nearby nodes */
				if (!matches.containsKey(node) && findNearestNodes(node, goldNodeList).size() == 0
						&& !isContinuation(node, goldNodeList, false)){
					if (writeDetails) System.out.println("Is Excess");
					excessNodes.put(node, excessWeight);
					// don't add to the weight if this node is the parent of a spur
					if (!spurList.contains(node)){
						weightSum += excessWeight;
					}
				}
				else{
					if (writeDetails) System.out.println("Not Excess");
					numExcessNodes = 0;
					excessWeight = 0;
				}
			}
			// Terminal node
			else{
				if (writeDetails) System.out.println("Is Termination");
				numExcessNodes = 0;
				excessWeight = 0;
				// Determine whether terminal node is excess (miss)
				// Excess if not a spur, no match, no match to parent, and no non-matched node within threshold
				if (!spurList.contains(node) && !matches.containsKey(node) && !matches.containsKey(node.getParent())
						&& findNearestNodes(node, goldNodeList).size() == 0){
					// See if an unselected gold node matches
					nearbyNodes = findNearestNodes(node, goldNodeList);
					if (nearbyNodes.size() > 0){
						// Remove nearest node so it cannot be used again
						removeNearestNode(node, nearbyNodes, goldNodeList);
					}
					else{
						if (writeDetails) System.out.println("Is Excess");
						numExcessNodes = 1;
						excessWeight = 1;
						excessNodes.put(node, 1);
						weightSum += excessWeight;
					}
				}
			}
			// Save excess nodes for display at end
			directTermExcess.put(node, numExcessNodes);
		}
		
		return weightSum;
	}
	
	private double getExcessWeight(double excess, BinaryTreeNode node, Map<BinaryTreeNode,Integer> excessMap){
		switch (weightMode){
		case WEIGHT_SQRT_DEGREE:
			return Math.sqrt(excess);
		case WEIGHT_DEGREE_HARMONIC_MEAN:
			return 2.0/(1.0/excessMap.get(node.getLeft())
					+ 1.0/excessMap.get(node.getRight()));
		case WEIGHT_UNIFORM:
			return (excess > 0) ? 1 : 0;
		case WEIGHT_DEGREE:
		default:
			return excess;
		}
	}
	
	private void removeNearestNode(BinaryTreeNode node, List<BinaryTreeNode> nearbyNodes, List<BinaryTreeNode> goldNodeList){
		double distance, closestDistance = -1;
		BinaryTreeNode closestMatch = null, nearbyNode;
		for (int j = 0; j < nearbyNodes.size(); j++){
			nearbyNode = nearbyNodes.get(j);
			// Check whether nodes' parents match

			distance = getDistance(node, nearbyNode);
			if (closestDistance == -1 || distance < closestDistance){
				closestDistance = distance;
				closestMatch = nearbyNode;
			}
		}
		goldNodeList.remove(closestMatch);
	}
	
	private void generateNodeWeights(List<SwcTreeNode> roots){
		for (int i = 0; i < roots.size(); i++){
			generateNodeWeights(roots.get(i));
		}		
	}
	private void generateNodeWeights(SwcTreeNode root){
		// load up weightMap by degree minus all terminal spurs
		BinaryTreeNode binaryRoot, node;
		LinkedStack initStack, mainStack;
		Map<BinaryTreeNode,Integer> degreeMap = new HashMap<BinaryTreeNode,Integer>();
		int degree;
		for (int i = 0; i < root.getChildren().size(); i++){
			binaryRoot = (BinaryTreeNode)root.getChildren().get(i);
			initStack = new LinkedStack();
			initStack.push(binaryRoot);
			mainStack = new LinkedStack();
			mainStack.push(binaryRoot);
			// Add all nodes to the main stack
			while (!initStack.isEmpty()){
				node = (BinaryTreeNode)initStack.pop();
				if (node.hasChildren()){
					initStack.push(node.getLeft());
					initStack.push(node.getRight());
					mainStack.push(node.getLeft());
					mainStack.push(node.getRight());
				}
			}
			
			if (isTestEnvironment()) System.out.println("Generating node weights");
			
			// Check for spurs using the mainStack, from terminals to root
			while (!mainStack.isEmpty()){
				node = (BinaryTreeNode)mainStack.pop();

				if (weightMode == WEIGHT_UNIFORM){
					weightMap.put(node, new Double(1));
				}
				// Calculate weight by determining each node's degree, taking spurs out of calculation
				else{
					writeDetails = isWithinXYThreshold(xyCheck, node.getSwcData());
					if (writeDetails || debug){
						System.out.println("Determining weight for "+node.getSwcData());
					}
					if (node.isLeaf()){
						if (writeDetails || debug) System.out.println("  Is leaf");
						weightMap.put(node, 1.0);
						degreeMap.put(node, 1);
					}
					else {
						degree = 0;
						if (node.getLeft().isLeaf() && node.getRight().isLeaf()){
							if (spurList.contains(node.getLeft()) 
									|| spurList.contains(node.getRight())){
								// If one child is spur and other not, this node is a continuation, 
								// but its weight still needs to be passed up to parent
								if (writeDetails || debug) System.out.println("  At least one leaf child is spur");
								degree = 1;
							}
							else{
								if (writeDetails || debug) System.out.println("  Both leaf children, neither spur");
								degree = 2;
							}
						}
						else if (spurList.contains(node)){
							if (writeDetails || debug) System.out.println("  One child bif, one spur leaf");
							// One spur leaf, one bifurcation
							if (node.getLeft().isLeaf()){
								// Right node is bifurcation with weight, pass it upwards
								degree += ((Integer)degreeMap.get(node.getRight())).intValue();
							}
							else{
								// Left node is bifurcation with weight, pass it upwards
								degree += ((Integer)degreeMap.get(node.getLeft())).intValue();
							}
						}
						else{
							if (writeDetails || debug) System.out.println("  Both children bifurcations");
							degree += ((Integer)degreeMap.get(node.getLeft())).intValue();
							degree += ((Integer)degreeMap.get(node.getRight())).intValue();
						}
						degreeMap.put(node, new Integer(degree));
						switch (weightMode){
						case WEIGHT_DEGREE:
							weightMap.put(node, new Double(degree));
							break;
						case WEIGHT_SQRT_DEGREE:
							weightMap.put(node, new Double(Math.sqrt(degree)));
							break;
						case WEIGHT_DEGREE_HARMONIC_MEAN:
							weightMap.put(node, new Double(
								2.0 / (1.0/degreeMap.get(node.getLeft()) +
									   1.0/degreeMap.get(node.getRight()))		
							));
							break;
						case WEIGHT_PATH_LENGTH:
							weightMap.put(node, new Double(node.getSwcData().getSecondaryData().getPathLength()));
							break;
						}
					}
				}
			}
		}
	}

	// Removes spurs from each of the trees in the entered set (the children of "root")
	private void removeSpurs(SwcTreeNode root, double threshold){
		List<BinaryTreeNode> children = new ArrayList<BinaryTreeNode>();
		BinaryTreeNode binaryRoot;
		for (int i = 0; i < root.getChildren().size(); i++){
			binaryRoot = (BinaryTreeNode)root.getChildren().get(i);
			binaryRoot = removeSpurs(binaryRoot, threshold);
			children.add(binaryRoot);
		}
		root.setChildren(children);
	}
	
	/**
	 * 
	 * @param root
	 * @param threshold
	 * @return Puts any terminal branches that are within the spur parameter setting into a list of spur nodes that
	 * will not count against the score. 
	 * 
	 */
	private BinaryTreeNode removeSpurs(
			BinaryTreeNode root, double threshold){
		BinaryTreeNode node, newNode;
		spurList = new ArrayList<BinaryTreeNode>();
		double distance;
		boolean bothChildrenAreSpurs;
		
		if (isTestEnvironment()) System.out.println("Removing Spurs");
		
		LinkedStack stack = new LinkedStack();
		// Make sure the root isn't also a termination
		if (root.hasChildren()){
			stack.push(root);
		}
		// Traverse the entire tree looking for spurs
		while (!stack.isEmpty()){
			node = (BinaryTreeNode)stack.pop();
			if (isWithinXYThreshold(xyCheck, node.getSwcData()))
				System.out.println("Spur check on "+getPositionString(node));
			if (node.hasChildren()){
				stack.push(node.getLeft());
				stack.push(node.getRight());
			}
			else{
				// Node is termination: determine if spur
				// If this is the test reconstruction and the node has been matched, don't count it as a spur
				distance = node.getSwcData().getSecondaryData().getPathLength();
				if (distance < threshold && !matches.containsKey(node)){
					if (isWithinXYThreshold(xyCheck, node.getSwcData()))
						System.out.println("Is Spur");
					bothChildrenAreSpurs = false;
					// Add to spurs list
					spurList.add(node);
					if (node.isRight()){
						// Node is right, left hasn't been checked yet
						newNode = node.getParent().getLeft();
						// Check to see if sibling is also spur
						if (!newNode.hasChildren() && newNode.getSwcData().getSecondaryData().getPathLength() < threshold
								 && !matches.containsKey(node)){
							if (isWithinXYThreshold(xyCheck, newNode.getSwcData()))
								System.out.println("Is Spur: "+getPositionString(newNode));
							bothChildrenAreSpurs = true;
							// Remove left node from stack (and add to spurs list)
							spurList.add(newNode);
							stack.pop();
						}
					}

					if (!bothChildrenAreSpurs){
						// Parent is not really a node now (is a continuation)
						spurList.add(node.getParent());
					}
				}
			}
		}
		return root;
	}
	
	/**
	 * 
	 * @param ancestorNode
	 * @param descendantNode
	 * @return EuclideanPoint that represents the location at which the gold standard path from the descendant node
	 * crosses its threshold boundary
	 */
	private EuclideanPoint getTrajectoryForPath(BinaryTreeNode ancestorNode, BinaryTreeNode descendantNode){
		LinkedStack pathStack = new LinkedStack();
		SwcDataNode ancestorData = ancestorNode.getSwcData();
		
		// Load nodes along path into a stack, from descendant up to ancestor
		while (descendantNode != ancestorNode){
			pathStack.push(descendantNode);
			descendantNode = descendantNode.getParent();
			if (descendantNode == null){
				// This should not happen if descendantNode is a descendant of ancestorNode
				return null;
			}
		}
		
		// Now we can start with the ancestor's immediate child and work down
		BinaryTreeNode nextDescendant;
		SwcSecondaryData secData;
		EuclideanPoint trajectory = new EuclideanPoint(), tmp;
		
		// Assumes original descendantNode != ancestorNode
		descendantNode = (BinaryTreeNode)pathStack.pop();
		boolean doneX = false, doneZ = false;
		
		while (!pathStack.isEmpty() && (!doneX || !doneZ)){
			secData = descendantNode.getSwcData().getSecondaryData();
			// Pull off next node down too to check which child trajectory to use
			nextDescendant = (BinaryTreeNode)pathStack.pop();
			
			// First see if descendantNode has a trajectory to nextDescendant
			if (nextDescendant.isLeft()){
				if (!doneX && secData.getLeftTrajectoryPoint().getX() != ReadSWC.TRAJECTORY_NONE){
					 trajectory.setX(secData.getLeftTrajectoryPoint().getX());
					 trajectory.setY(secData.getLeftTrajectoryPoint().getY());
					 doneX = true;
				}
				if (!doneZ && secData.getLeftTrajectoryPoint().getZ() != ReadSWC.TRAJECTORY_NONE){
					 trajectory.setZ(secData.getLeftTrajectoryPoint().getZ());
					 doneZ = true;
				}
			}
			else if (nextDescendant.isRight() && secData.getRightTrajectoryPoint() != null){
				if (!doneX && secData.getRightTrajectoryPoint().getX() != ReadSWC.TRAJECTORY_NONE){
					 trajectory.setX(secData.getRightTrajectoryPoint().getX());
					 trajectory.setY(secData.getRightTrajectoryPoint().getY());
					 doneX = true;
				}
				if (!doneZ && secData.getRightTrajectoryPoint().getZ() != ReadSWC.TRAJECTORY_NONE){
					 trajectory.setZ(secData.getRightTrajectoryPoint().getZ());
					 doneZ = true;
				}
			}
			// Otherwise descendant and next are close;

			// If threshold distance has been reached from ancestor, calculate point between descendants
			if (!doneX && !isWithinXYThreshold(ancestorData, nextDescendant.getSwcData())){
				tmp = ReadSWC.calculateTrajectoryXY(ancestorData, descendantNode.getSwcData(), nextDescendant.getSwcData(),XYThreshold);
				trajectory.setX(tmp.getX());
				doneX = true;
			}
			if (!doneZ && (Math.abs(ancestorData.getZ() - nextDescendant.getSwcData().getZ()) > ZThreshold)){
				trajectory.setZ(ReadSWC.calculateTrajectoryZ(ancestorData, descendantNode.getSwcData(), nextDescendant.getSwcData(),ZThreshold));
				doneZ = true;
			}
			
			// Move down by setting nextDescendant as current one for trajectory check
			descendantNode = nextDescendant;
		}
		
		if (!doneX){
			// Distance from ancestor to descendant is within threshold distance
			// The last descendantNode should be the descendantNode specified in the input parameters
			trajectory.setX(descendantNode.getSwcData().getX());
			trajectory.setY(descendantNode.getSwcData().getY());
		}
		if (!doneZ){
			trajectory.setZ(descendantNode.getSwcData().getZ());
		}
		
		return trajectory;
	}
	
	private double getEndNodeXYDistanceDifference(SwcDataNode goldData, EuclideanPoint trajectory, SwcDataNode testData){
		double goldDist = SwcDataUtils.getXYDistance(goldData, trajectory);
		double testDist = SwcDataUtils.getXYDistance(testData, trajectory);
		return goldDist - testDist;
	}

	private double getEndNodeZDistanceDifference(SwcDataNode goldData, EuclideanPoint trajectory, SwcDataNode testData){
		double goldDist = Math.abs(goldData.getZ() - trajectory.getZ());
		double testDist = Math.abs(testData.getZ() - trajectory.getZ());
		return goldDist - testDist;
	}

	private double getDistance(BinaryTreeNode node1, BinaryTreeNode node2){
		return SwcDataUtils.getDistance(node1.getSwcData(), node2.getSwcData());
	}
	private double getXYDistance(BinaryTreeNode node1, BinaryTreeNode node2){
		return SwcDataUtils.getXYDistance(node1.getSwcData(), node2.getSwcData());
	}
	private double getZDistance(BinaryTreeNode node1, BinaryTreeNode node2){
		return Math.abs(node1.getSwcData().getZ() - node2.getSwcData().getZ());
	}
	private boolean isWithinThreshold(BinaryTreeNode node1, BinaryTreeNode node2){
		return isWithinThreshold(node1.getSwcData(), node2.getSwcData());
	}
	private boolean isWithinThreshold(SwcDataNode node1, SwcDataNode node2){
		return (SwcDataUtils.getXYDistance(node1, node2) <= XYThreshold
				&& Math.abs(node1.getZ() - node2.getZ()) <= (ZThreshold+0.1));
		// ZThreshold increased fractionally to avoid potential floating point error
	}
	private boolean isWithinXYThreshold(EuclideanPoint point1, EuclideanPoint point2){
		return (SwcDataUtils.getXYDistance(point1, point2) <= XYThreshold);
	}
	private boolean isWithinDistantMatchThreshold(BinaryTreeNode node1, BinaryTreeNode node2){
		return isWithinDistantMatchThreshold(node1.getSwcData(), node2.getSwcData());
	}
	private boolean isWithinDistantMatchThreshold(SwcDataNode node1, SwcDataNode node2){
		return (SwcDataUtils.getXYDistance(node1, node2) <= XYThreshold * 3
				&& Math.abs(node1.getZ() - node2.getZ()) <= ZThreshold * 3 + 0.1);
	}
	
	private String getPositionString(EuclideanPoint node){
		NumberFormat fmt = NumberFormat.getInstance();
		fmt.setMaximumFractionDigits(3);
		fmt.setGroupingUsed(false);
		if (scaleZ != 0){
			return "("+fmt.format(node.getX())+","+fmt.format(node.getY())+","+fmt.format(node.getZ()/scaleZ)+")";
		}
		else{
			return "("+fmt.format(node.getX())+","+fmt.format(node.getY())+")";
		}
	}
	private String getPositionString(BinaryTreeNode node){
		NumberFormat fmt = NumberFormat.getInstance();
		fmt.setMaximumFractionDigits(3);
		fmt.setGroupingUsed(false);
		if (scaleZ != 0){
			return "("+fmt.format(node.getSwcData().getX())+","+fmt.format(node.getSwcData().getY())+","+fmt.format(node.getSwcData().getZ()/scaleZ)+")";
		}
		else{
			return "("+fmt.format(node.getSwcData().getX())+","+fmt.format(node.getSwcData().getY())+")";
		}
	}

	public List<SwcDataNode> getOffsets(int index){
		List<SwcDataNode> offsetsCopy = new ArrayList<SwcDataNode>(offsets.size());
		for (int i = 0; i < offsetsCopy.size(); i++){
			offsetsCopy.add(getOffset(i));
		}
		return offsetsCopy;
	}

	public SwcDataNode getOffset(int index){
		SwcDataNode offset = offsets.get(index);
		SwcDataNode offsetCopy = new SwcDataNode();
		offsetCopy.setX(offset.getX());
		offsetCopy.setY(offset.getY());
		offsetCopy.setZ(offset.getZ());
		return offsetCopy;
	}
	public double getOffsetX(int index){
		return offsets.size() > index ? offsets.get(index).getX() : null;
	}
	public double getOffsetY(int index){
		return offsets.size() > index ? offsets.get(index).getY() : null;
	}
	public double getOffsetZ(int index){
		return offsets.size() > index ? offsets.get(index).getZ() : null;
	}

	/**
	 * 
	 * @param sampleRoot
	 * @param baselineRoot
	 * Updates nodes in sampleRoot's tree based on position offset between sampleRoot and baselineRoot
	 * 
	 */
	private void alignRoots(SwcTreeNode testRoot, SwcTreeNode goldStandardRoot){
		// Determine translation offset if any
		SwcDataNode offset = determineOffset(testRoot, goldStandardRoot);
		offsets.add(offset);
		LinkedStack stack;
		SwcDataNode data;
		BinaryTreeNode node;

		System.out.println("Position translation to align test tree with gold standard:\n"+offset.getX()+","+offset.getY()+","+offset.getZ());
		
		// Translate sample root node
		testRoot.setX(testRoot.getX() + offset.getX());
		testRoot.setY(testRoot.getY() + offset.getY());
		testRoot.setZ(testRoot.getZ() + offset.getZ());

		// Translate all sample nodes
		for (int i = 0; i < testRoot.getChildren().size(); i++){
			stack = new LinkedStack();
			node = (BinaryTreeNode)testRoot.getChildren().get(i);
			stack.push(node);
			
			while(!stack.isEmpty()){
				node = (BinaryTreeNode)stack.pop();
				data = (SwcDataNode)node.getData();
				data.setX(data.getX() + offset.getX());
				data.setY(data.getY() + offset.getY());
				data.setZ(data.getZ() + offset.getZ());
				
				if (node.hasChildren()){
					stack.push(node.getRight());
					stack.push(node.getLeft());
				}
			}
		}
		if (debug) System.out.println("Done aligning test tree with gold standard");
	}

	private static SwcDataNode determineOffset(SwcTreeNode sampleRoot, SwcTreeNode baselineRoot){
		SwcDataNode sampleData = sampleRoot.getSwcData();
		SwcDataNode baselineData = baselineRoot.getSwcData();
		// If each data set only has one tree, align based on first bifurcation
		/*if (sampleRoot.getChildren().size() == 1 && 
				baselineRoot.getChildren().size() == 1){
			sampleData = (SwcDataNode)((BinaryTreeNode)sampleRoot.getChildren().iterator().next()).getData();
			baselineData = (SwcDataNode)((BinaryTreeNode)baselineRoot.getChildren().iterator().next()).getData();
		}*/
		SwcDataNode offset = new SwcDataNode();
		offset.setX(baselineData.getX() - sampleData.getX());
		offset.setY(baselineData.getY() - sampleData.getY());
		offset.setZ(baselineData.getZ() - sampleData.getZ());
		return offset;
	}

	// Sets the available program parameters
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
    	opt = new FlaggedOption(PARAM_TEST_DATA)
			.setStringParser(JSAP.STRING_PARSER)
			.setRequired(true)
			.setShortFlag('T') 
			.setLongFlag(PARAM_TEST_DATA);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_GOLD_STANDARD_DATA)
			.setStringParser(JSAP.STRING_PARSER)
			.setRequired(true)
			.setShortFlag('G') 
			.setLongFlag(PARAM_GOLD_STANDARD_DATA);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_DATASET)
			.setStringParser(JSAP.INTEGER_PARSER)
	    	.setRequired(true)
			.setShortFlag('D') 
			.setLongFlag(PARAM_DATASET);
        opt.setHelp("            1: Cerebellar Climbing Fibers \n" +
					 "            2: Hippocampal CA3 Interneurons \n" +
					 "            3: Neocortical Layer 1 Axons \n" +
					 "            4: Neuromuscular Projection Fibers \n" +
					 "            5: Olfactory Projection Fibers \n"+
					 "			  6: Visual Cortical Layer 6 Neuron");
	    if (mode == MODE_TEST_ENVIRONMENT || mode == MODE_OPEN){
	    	opt.setRequired(false);
	    	opt.setDefault("0");
	    }

	    jsap.registerParameter(opt);

		opt = new FlaggedOption(PARAM_LIST_MISSES)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("false")
			.setShortFlag('m') 
			.setLongFlag(PARAM_LIST_MISSES);
		jsap.registerParameter(opt);

		/** These parameters are only available when using the test version or the fully open environment 
		 * (as opposed to the DIADEM competition environment built for the 5 specific data sets)
		 */
	    if (mode == MODE_TEST_ENVIRONMENT || mode == MODE_OPEN){
		    opt = new FlaggedOption(PARAM_WEIGHT_SCHEME)
				.setStringParser(JSAP.INTEGER_PARSER)
				.setRequired(false)
				.setDefault(""+WEIGHT_DEGREE)
				.setShortFlag('w') 
				.setLongFlag(PARAM_WEIGHT_SCHEME);
	        opt.setHelp("            1: Weight by Degree \n" +
					 "            2: Uniform Weight \n" +
					 "            3: Sqrt of Degree \n" +
					 "            4: Harmonic mean of node's subtrees' degree \n");
		    jsap.registerParameter(opt);
	
		    opt = new FlaggedOption(PARAM_XY_THRESHOLD)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault("1")
				.setShortFlag('x') 
				.setLongFlag(PARAM_XY_THRESHOLD);
		    jsap.registerParameter(opt);

		    opt = new FlaggedOption(PARAM_XY_Z_RATIO)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault("1")
				.setShortFlag('R') 
				.setLongFlag(PARAM_XY_Z_RATIO);
		    opt.setHelp(" The XY/Z Ratio is the number of XY pixels per Z image");
		    jsap.registerParameter(opt);

		    opt = new FlaggedOption(PARAM_Z_THRESHOLD)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault("1")
				.setShortFlag('z') 
				.setLongFlag(PARAM_Z_THRESHOLD);
	        opt.setHelp(" Z Threshold parameter unit is images. It is converted to pixels using the XY pixels per Z image ratio");
		    jsap.registerParameter(opt);

		    opt = new FlaggedOption(PARAM_TERMINAL_THRESHOLD)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault(""+DEFAULT_TERMINAL_THRESHOLD)
				.setShortFlag('t') 
				.setLongFlag(PARAM_TERMINAL_THRESHOLD);
		    jsap.registerParameter(opt);
		    opt.setHelp("Threshold Euclidean distance from which a terminal node may be matched");
		    
		    opt = new FlaggedOption(PARAM_TERMINAL_PATH_THRESHOLD)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault(""+DEFAULT_TERMINAL_PATH_THRESHOLD)
				.setShortFlag('p') 
				.setLongFlag(PARAM_TERMINAL_PATH_THRESHOLD);
		    jsap.registerParameter(opt);
	        opt.setHelp("Maximum difference in path length allowed between gold standard termination and matched test node (after accounting for path length error).\n"+
	        		"Defaults to the terminal Euclidean distance threshold (--terminal-threshold).");

		    opt = new FlaggedOption(PARAM_XY_PATH_THRESHOLD)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault("1")
				.setShortFlag(JSAP.NO_SHORTFLAG) 
				.setLongFlag(PARAM_XY_PATH_THRESHOLD);
		    jsap.registerParameter(opt);
		
		    opt = new FlaggedOption(PARAM_Z_PATH_THRESHOLD)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault("1")
				.setShortFlag(JSAP.NO_SHORTFLAG) 
				.setLongFlag(PARAM_Z_PATH_THRESHOLD);
		    jsap.registerParameter(opt);

			opt = new FlaggedOption(PARAM_REMOVE_SPURS)
				.setStringParser(JSAP.DOUBLE_PARSER)
				.setRequired(false)
				.setDefault("0")
				.setShortFlag('r') 
				.setLongFlag(PARAM_REMOVE_SPURS);
			jsap.registerParameter(opt);

		    opt = new FlaggedOption(PARAM_LIST_CONTINUATIONS)
				.setStringParser(JSAP.BOOLEAN_PARSER)
				.setRequired(false)
				.setDefault("false")
				.setShortFlag('c') 
				.setLongFlag(PARAM_LIST_CONTINUATIONS);
			jsap.registerParameter(opt);
	
			opt = new FlaggedOption(PARAM_LIST_DISTANT_MATCHES)
				.setStringParser(JSAP.BOOLEAN_PARSER)
				.setRequired(false)
				.setDefault("false")
				.setShortFlag('d') 
				.setLongFlag(PARAM_LIST_DISTANT_MATCHES);
			jsap.registerParameter(opt);
	
			opt = new FlaggedOption(PARAM_Z_PATH)
				.setStringParser(JSAP.BOOLEAN_PARSER)
				.setRequired(false)
				.setDefault("false")
				.setShortFlag(JSAP.NO_SHORTFLAG) 
				.setLongFlag(PARAM_Z_PATH);
			jsap.registerParameter(opt);
	
			opt = new FlaggedOption(PARAM_COUNT_EXCESS_NODE)
				.setStringParser(JSAP.BOOLEAN_PARSER)
				.setRequired(false)
				.setDefault("true")
				.setShortFlag(JSAP.NO_SHORTFLAG) 
				.setLongFlag(PARAM_COUNT_EXCESS_NODE);
			jsap.registerParameter(opt);

	        opt = new FlaggedOption(PARAM_MICRONS)
				.setStringParser(JSAP.BOOLEAN_PARSER)
				.setRequired(false)
				.setDefault("false")
				.setShortFlag('M') 
				.setLongFlag(PARAM_MICRONS);
		    jsap.registerParameter(opt);

		    /* Parameter for functionality being considered - simply adjusts test reconstruction coordinates so root nodes match
			opt = new FlaggedOption(PARAM_ALIGN_TREES_BY_ROOT)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("false")
			.setShortFlag('a') 
			.setLongFlag(PARAM_ALIGN_TREES_BY_ROOT);
			jsap.registerParameter(opt);
			*/

			if (mode == MODE_TEST_ENVIRONMENT){
				opt = new FlaggedOption("X")
					.setStringParser(JSAP.DOUBLE_PARSER)
					.setRequired(false)
					.setDefault("0")
					.setShortFlag('X')
					.setLongFlag("X");
				jsap.registerParameter(opt);
		
				opt = new FlaggedOption("Y")
					.setStringParser(JSAP.DOUBLE_PARSER)
					.setRequired(false)
					.setDefault("0")
					.setShortFlag('Y')
					.setLongFlag("Y");
				jsap.registerParameter(opt);
		
				opt = new FlaggedOption("debug")
					.setStringParser(JSAP.BOOLEAN_PARSER)
					.setRequired(false)
					.setDefault("false")
					.setShortFlag(JSAP.NO_SHORTFLAG)
					.setLongFlag("debug");
				jsap.registerParameter(opt);
			}

	    }

		JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java -jar DiademMetric.jar");
            System.err.println("                " + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

	public static void main(String args[]){
		try{
			// Read in command line parameters
			JSAPResult config = readParameters(args);
			String testSwcFilename = config.getString(PARAM_TEST_DATA);
			String goldSwcFilename = config.getString(PARAM_GOLD_STANDARD_DATA);
			int dataset = config.getInt(PARAM_DATASET);
			File testSwcFile = new File(testSwcFilename);
			File goldSwcFile = new File(goldSwcFilename);

			// Validate parameters
			validate(goldSwcFilename, testSwcFilename, dataset);

			// Create DiademMetric object and assign any relevant parameters
			DiademMetric metric = new DiademMetric(testSwcFile,goldSwcFile,dataset);
			metric.setListMisses(config.getBoolean(PARAM_LIST_MISSES));

			// Read parameters that are available in the test or open modes
			if (isTestEnvironment() || mode == MODE_OPEN){
				if (config.userSpecified(PARAM_MICRONS)){
					metric.setMicrons(config.getBoolean(PARAM_MICRONS));
				}
				if (config.userSpecified(PARAM_XY_THRESHOLD)){
					metric.setXYThreshold(config.getDouble(PARAM_XY_THRESHOLD));
				}
				if (config.userSpecified(PARAM_Z_THRESHOLD)){
					metric.setRawZThreshold(config.getDouble(PARAM_Z_THRESHOLD));
				}
				if (config.userSpecified(PARAM_XY_Z_RATIO)){
					metric.setScaleZ(config.getDouble(PARAM_XY_Z_RATIO));
				}
				if (config.userSpecified(PARAM_TERMINAL_THRESHOLD)){
					metric.setTerminalThreshold(config.getDouble(PARAM_TERMINAL_THRESHOLD));
				}
				if (config.userSpecified(PARAM_TERMINAL_PATH_THRESHOLD)){
					metric.setTerminalPathThreshold(config.getDouble(PARAM_TERMINAL_PATH_THRESHOLD));
				}
				else{
					metric.setTerminalPathThreshold(metric.getTerminalThreshold());
				}
				if (config.userSpecified(PARAM_XY_PATH_THRESHOLD)){
					metric.setXYPathErrorThreshold(config.getDouble(PARAM_XY_PATH_THRESHOLD));
				}
				if (config.userSpecified(PARAM_Z_PATH_THRESHOLD)){
					metric.setZPathErrorThreshold(config.getDouble(PARAM_Z_PATH_THRESHOLD));
				}
				if (config.userSpecified(PARAM_WEIGHT_SCHEME)){
					metric.setWeightMode(config.getInt(PARAM_WEIGHT_SCHEME));
				}
				if (config.userSpecified(PARAM_Z_PATH)){
					metric.setZPath(config.getBoolean(PARAM_Z_PATH));
				}
				if (config.userSpecified(PARAM_REMOVE_SPURS)){
					metric.setRemoveSpurs(config.getDouble(PARAM_REMOVE_SPURS));
				}
				if (config.userSpecified(PARAM_COUNT_EXCESS_NODE)){
					metric.setCountExcessNodes(config.getBoolean(PARAM_COUNT_EXCESS_NODE));
				}
				//metric.setAlignTreesByRoot(config.getBoolean(PARAM_ALIGN_TREES_BY_ROOT));
				metric.setListContinuations(config.getBoolean(PARAM_LIST_CONTINUATIONS));
				metric.setListDistantMatches(config.getBoolean(PARAM_LIST_DISTANT_MATCHES));
				if (mode == MODE_TEST_ENVIRONMENT){
					metric.setX(config.getDouble("X"));
					metric.setY(config.getDouble("Y"));
					metric.setDebug(config.getBoolean("debug"));
				}
			}
			
			// Run the metric
			metric.scoreReconstruction();
		} catch (DataFormatException e){
			System.out.println("Failure due to: "+e.getMessage());
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	// Validate certain input parameters
	private static void validate(String goldSwcFilename, String testSwcFilename, int dataset){
		File testSwcFile = new File(testSwcFilename);
		File goldSwcFile = new File(goldSwcFilename);

		boolean stop = false;
		if (!testSwcFile.exists()){
			System.out.println("Test file "+testSwcFilename+" does not exist");
			stop = true;
		}
		else if (!testSwcFile.isDirectory() && !testSwcFilename.endsWith(".swc") && !testSwcFilename.endsWith(".SWC")){
			System.out.println("Test file "+testSwcFilename+" is not an SWC file or directory");
			stop = true;
		}
		if (!goldSwcFile.exists()){
			System.out.println("Gold standard file/directory "+goldSwcFilename+" does not exist");
			stop = true;
		}
		else if (!goldSwcFile.isDirectory() && !goldSwcFilename.endsWith(".swc") && !goldSwcFilename.endsWith(".SWC")){
			System.out.println("Gold standard file/directory "+goldSwcFilename+" is not an SWC file or directory");
			stop = true;
		}
		if (!stop &&
				((goldSwcFile.isDirectory() && !testSwcFile.isDirectory()) ||
						(!goldSwcFile.isDirectory() && testSwcFile.isDirectory()))){
			System.out.println("Gold standard and test parameters must both be swc files or both be directories");
			stop = true;
		}
		if (mode != MODE_OPEN && (dataset < 1 || dataset > 6)){
			System.out.println("Data set number invalid, please use one of the following: \n" +
					"            1: Cerebellar Climbing Fibers \n" +
					"            2: Hippocampal CA3 Interneurons \n" +
					"            3: Neocortical Layer 6 Axons \n" +
					"            4: Neuromuscular Projection Fibers \n" +
					"            5: Olfactory Projection Fibers \n" +
					"            6: Visual Cortical Layer 6 Neuron");
			stop = true;
		}

		if (stop){
			System.exit(0);
		}

	}
	
}
